from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, Optional, Union, Dict, Any

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from qcarchivetesting.helpers import read_procedure_data, read_record_data, find_test_data
from qcelemental.models import (
    Molecule,
    FailedOperation,
    ComputeError,
    OptimizationResult as QCEl_OptimizationResult,
)

from qcfractal.components.optimization.record_db_models import OptimizationRecordORM
from qcfractalcompute.compress import compress_result
from qcportal.optimization import OptimizationSpecification, OptimizationRecord
from qcportal.record_models import PriorityEnum, RecordStatusEnum
from qcportal.singlepoint import QCSpecification, SinglepointProtocols
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcportal.managers import ManagerName


all_test_data = find_test_data("opt_*")
all_includes = ["initial_molecule", "final_molecule", "trajectory", "molecule"]

test_specs = [
    OptimizationSpecification(
        program="optprog1",
        keywords={},
        protocols={"trajectory": "initial_and_final"},
        qc_specification=QCSpecification(
            program="prog1",
            driver="deferred",
            method="b3lyp",
            basis="6-31G*",
            keywords={"k": "value"},
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    ),
    OptimizationSpecification(
        program="optprog2",
        keywords={"k": "v"},
        protocols={"trajectory": "none"},
        qc_specification=QCSpecification(
            program="Prog2",
            driver="deferred",
            method="Hf",
            basis="def2-TZVP",
            keywords={"k": "v"},
        ),
    ),
    OptimizationSpecification(
        program="optPRog3",
        keywords={"k2": "v2"},
        qc_specification=QCSpecification(
            program="Prog3",
            driver="deferred",
            method="pbe0",
            basis="",
            keywords={"o": 1, "v": 2.123},
            protocols=SinglepointProtocols(stdout=False, wavefunction="orbitals_and_eigenvalues"),
        ),
    ),
    OptimizationSpecification(
        program="OPTPROG4",
        qc_specification=QCSpecification(
            program="ProG4",
            driver="deferred",
            method="pbe",
            basis=None,
            protocols=SinglepointProtocols(stdout=False, wavefunction="return_results"),
        ),
    ),
]


def load_procedure_data(name: str) -> Tuple[OptimizationSpecification, Molecule, QCEl_OptimizationResult]:
    data = read_procedure_data(name)

    return (
        pydantic.parse_obj_as(OptimizationSpecification, data["specification"]),
        pydantic.parse_obj_as(Molecule, data["initial_molecule"]),
        pydantic.parse_obj_as(QCEl_OptimizationResult, data["result"]),
    )


def load_record_data(name: str) -> OptimizationRecord:
    data = read_record_data(name)
    return OptimizationRecord(**data)


def submit_procedure_data(
    storage_socket: SQLAlchemySocket,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
) -> Tuple[int, QCEl_OptimizationResult]:
    input_spec, molecule, result = load_procedure_data(name)
    meta, record_ids = storage_socket.records.optimization.add(
        [molecule], input_spec, compute_tag, compute_priority, None, True
    )
    assert meta.success
    assert len(record_ids) == 1
    assert meta.n_inserted == 1

    return record_ids[0], result


def run_procedure_data(
    storage_socket: SQLAlchemySocket,
    manager_name: ManagerName,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
    end_status: RecordStatusEnum = RecordStatusEnum.complete,
):
    time_0 = now_at_utc()
    record_id, result = submit_procedure_data(storage_socket, name, compute_tag, compute_priority)
    time_1 = now_at_utc()

    with storage_socket.session_scope() as session:
        record = session.get(OptimizationRecordORM, record_id)
        assert record.status == RecordStatusEnum.waiting

    if end_status == RecordStatusEnum.error:
        result = FailedOperation(
            error=ComputeError(error_type="test_error", error_message="this is just a test error"),
        )

    manager_programs = storage_socket.managers.get([manager_name.fullname])[0]["programs"]
    tasks = storage_socket.tasks.claim_tasks(manager_name.fullname, manager_programs, [compute_tag], limit=100)
    assert len(tasks) == 1

    result_compressed = compress_result(result.dict())
    result_dict = {tasks[0]["id"]: result_compressed}
    storage_socket.tasks.update_finished(manager_name.fullname, result_dict)

    time_2 = now_at_utc()

    with storage_socket.session_scope() as session:
        record = session.get(OptimizationRecordORM, record_id)
        assert record.status == end_status
        assert time_0 < record.created_on < time_1
        assert time_1 < record.modified_on < time_2
        assert time_1 < record.compute_history[0].modified_on < time_2

    return record_id


def compare_optimization_specs(
    input_spec: Union[OptimizationSpecification, Dict[str, Any]],
    output_spec: Union[OptimizationSpecification, Dict[str, Any]],
) -> bool:
    if isinstance(input_spec, dict):
        input_spec = OptimizationSpecification(**input_spec)
    if isinstance(output_spec, dict):
        output_spec = OptimizationSpecification(**output_spec)

    return input_spec == output_spec



"""Add dataset record count table

Revision ID: 39abe82f389f
Revises: 83127ad1a00e
Create Date: 2026-03-01 09:22:02.306484

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "39abe82f389f"
down_revision = "83127ad1a00e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "dataset_record_count",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("record_count", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["dataset_id"], ["base_dataset.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("dataset_id"),
    )

    # Populate the table with initial record counts
    op.execute(sa.text("""
                       INSERT INTO dataset_record_count (dataset_id, record_count)
                            (SELECT dataset_id, COUNT(*) FROM dataset_direct_records_view GROUP BY dataset_id)
                       """))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("dataset_record_count")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import (
    Column,
    Integer,
    JSON,
    ForeignKey,
    String,
    Boolean,
    Index,
    UniqueConstraint,
    CheckConstraint,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship

from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.db_socket import BaseORM, PlainMsgpackExt

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class ServiceDependencyORM(BaseORM):
    """
    Table for storing dependencies of a service

    These are other records that are required to be completed before
    a service will iterate.
    """

    __tablename__ = "service_dependency"

    id = Column(Integer, primary_key=True)

    service_id = Column(Integer, ForeignKey("service_queue.id", ondelete="cascade"), nullable=False)
    record_id = Column(Integer, ForeignKey(BaseRecordORM.id), nullable=False)
    extras = Column(JSONB, nullable=False)

    record = relationship(BaseRecordORM)

    # We make extras part of the unique constraint because rarely the same dependency will be
    # submitted but with different extras (position, etc)
    __table_args__ = (UniqueConstraint("service_id", "record_id", "extras", name="ux_service_dependency"),)

    _qcportal_model_excludes = ["id", "service_id"]


class ServiceQueueORM(BaseORM):
    """
    Table for storing service information
    """

    __tablename__ = "service_queue"

    id = Column(Integer, primary_key=True)

    record_id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), nullable=False)
    record = relationship(BaseRecordORM, back_populates="service", uselist=False)

    compute_tag = Column(String, nullable=False)
    compute_priority = Column(Integer, nullable=False)
    find_existing = Column(Boolean, nullable=False)

    service_state = Column(PlainMsgpackExt)

    dependencies = relationship(
        ServiceDependencyORM, lazy="selectin", cascade="all, delete-orphan", passive_deletes=True
    )

    __table_args__ = (
        UniqueConstraint("record_id", name="ux_service_queue_record_id"),
        Index("ix_service_queue_compute_tag", "compute_tag"),
        CheckConstraint("compute_tag = LOWER(compute_tag)", name="ck_service_queue_compute_tag_lower"),
    )

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        # TODO - DEPRECATED - remove
        d["tag"] = d.pop("compute_tag")
        d["priority"] = d.pop("compute_priority")

        return d


class ServiceSubtaskRecordORM(BaseRecordORM):
    """
    Table for storing records associated with service iterations
    """

    __tablename__ = "service_subtask_record"

    id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), primary_key=True)

    # In other records, required_programs is a property. But here, it's basically
    # specified by whoever creates the record, so it's a column
    required_programs = Column(JSON, nullable=False)

    function = Column(String, nullable=False)

    function_kwargs = Column(JSON, nullable=False)
    results = Column(JSON, nullable=True)

    # We need the inherit_condition because we have two foreign keys to the BaseRecordORM and
    # we need to disambiguate
    __mapper_args__ = {
        "polymorphic_identity": "servicesubtask",
        "inherit_condition": (id == BaseRecordORM.id),
    }

    @property
    def short_description(self) -> str:
        progs = ",".join(self.required_programs.keys())
        return f"{self.function} [{progs}]"



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Tuple, Optional, Dict, List, Union, Any

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from qcelemental.models import Molecule, FailedOperation, ComputeError, OptimizationResult as QCEl_OptimizationResult

from qcarchivetesting.helpers import read_procedure_data, read_record_data, find_test_data
from qcfractal.components.torsiondrive.record_db_models import TorsiondriveRecordORM
from qcfractal.testing_helpers import run_service
from qcportal.optimization import OptimizationSpecification, OptimizationProtocols
from qcportal.record_models import PriorityEnum, RecordStatusEnum, RecordTask
from qcportal.singlepoint import SinglepointProtocols, QCSpecification
from qcportal.torsiondrive import TorsiondriveSpecification, TorsiondriveKeywords, TorsiondriveRecord
from qcportal.utils import recursive_normalizer

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcportal.managers import ManagerName

all_test_data = find_test_data("td_*")
all_includes = ["initial_molecules", "optimizations", "initial_molecule", "final_molecule"]

test_specs = [
    TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(1, 2, 3, 4)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        ),
    ),
    TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(7, 2, 9, 4), (5, 11, 3, 10)],
            grid_spacing=[30, 45],
            dihedral_ranges=[[-90, 90], [0, 180]],
            energy_decrease_thresh=1.0,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all", stdout=False),
            ),
        ),
    ),
]


def compare_torsiondrive_specs(
    input_spec: Union[TorsiondriveSpecification, Dict[str, Any]],
    output_spec: Union[TorsiondriveSpecification, Dict[str, Any]],
) -> bool:
    if isinstance(input_spec, dict):
        input_spec = TorsiondriveSpecification(**input_spec)
    if isinstance(output_spec, dict):
        output_spec = TorsiondriveSpecification(**output_spec)

    return input_spec == output_spec


def generate_task_key(task: RecordTask):
    # task is an optimization
    inp_data = task.function_kwargs["input_data"]
    assert inp_data["schema_name"] == "qcschema_optimization_input"

    mol_hash = inp_data["initial_molecule"]["identifiers"]["molecule_hash"]
    constraints = inp_data["keywords"].get("constraints", None)

    # Lookups may depend on floating point values
    constraints = recursive_normalizer(constraints)

    # This is the key in the dictionary of optimization results
    constraints_str = json.dumps(constraints, sort_keys=True)
    return mol_hash + "|" + constraints_str


def load_procedure_data(
    name: str,
) -> Tuple[TorsiondriveSpecification, List[Molecule], Dict[str, QCEl_OptimizationResult]]:
    data = read_procedure_data(name)

    return (
        pydantic.parse_obj_as(TorsiondriveSpecification, data["specification"]),
        pydantic.parse_obj_as(List[Molecule], data["initial_molecules"]),
        pydantic.parse_obj_as(Dict[str, QCEl_OptimizationResult], data["results"]),
    )


def load_record_data(name: str) -> TorsiondriveRecord:
    data = read_record_data(name)
    return TorsiondriveRecord(**data)


def submit_procedure_data(
    storage_socket: SQLAlchemySocket,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
) -> Tuple[int, Dict[str, QCEl_OptimizationResult]]:
    input_spec, molecules, result = load_procedure_data(name)
    meta, record_ids = storage_socket.records.torsiondrive.add(
        [molecules], input_spec, True, compute_tag, compute_priority, None, True
    )
    assert meta.success
    assert len(record_ids) == 1
    assert meta.n_inserted == 1

    return record_ids[0], result


def run_procedure_data(
    storage_socket: SQLAlchemySocket,
    manager_name: ManagerName,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
    end_status: RecordStatusEnum = RecordStatusEnum.complete,
):
    record_id, result = submit_procedure_data(storage_socket, name, compute_tag, compute_priority)

    with storage_socket.session_scope() as session:
        record = session.get(TorsiondriveRecordORM, record_id)
        assert record.status == RecordStatusEnum.waiting

    if end_status == RecordStatusEnum.error:
        failed_op = FailedOperation(
            error=ComputeError(error_type="test_error", error_message="this is just a test error"),
        )
        result = {x: failed_op for x in result}

    finished, n_optimizations = run_service(storage_socket, manager_name, record_id, generate_task_key, result, 200)
    assert finished

    with storage_socket.session_scope() as session:
        record = session.get(TorsiondriveRecordORM, record_id)
        assert record.status == end_status

    return record_id



from __future__ import annotations

from typing import TYPE_CHECKING

from qcportal.serverinfo import ErrorLogQueryFilters
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake


def test_serverinfo_socket_save_error(secure_snowflake: QCATestingSnowflake):
    storage_socket = secure_snowflake.get_storage_socket()
    admin_id = storage_socket.users.get("admin_user")["id"]
    read_id = storage_socket.users.get("read_user")["id"]

    userid_map = {admin_id: "admin_user", read_id: "read_user"}

    error_data_1 = {
        "error_text": "This is a test error",
        "user_id": admin_id,
        "request_path": "/api/v1/molecule",
        "request_headers": "fake_headers",
        "request_body": "fake body",
    }

    error_data_2 = {
        "error_text": "This is another test error",
        "user_id": read_id,
        "request_path": "/api/v1/molecule",
        "request_headers": "fake_headers",
        "request_body": "fake body",
    }

    all_errors = [error_data_1, error_data_2]
    id_1 = storage_socket.serverinfo.save_error(error_data_1)
    time_12 = now_at_utc()
    id_2 = storage_socket.serverinfo.save_error(error_data_2)

    errors = storage_socket.serverinfo.query_error_log(ErrorLogQueryFilters())
    assert len(errors) == 2

    # Returned in chrono order, newest first
    assert errors[0]["id"] == id_2
    assert errors[1]["id"] == id_1
    assert errors[0]["error_date"] > errors[1]["error_date"]

    for in_err, db_err in zip(reversed(all_errors), errors):
        assert in_err["error_text"] == db_err["error_text"]
        assert userid_map[in_err["user_id"]] == db_err["user"]
        assert in_err["request_path"] == db_err["request_path"]
        assert in_err["request_headers"] == db_err["request_headers"]
        assert in_err["request_body"] == db_err["request_body"]

    # Query by id
    err = storage_socket.serverinfo.query_error_log(ErrorLogQueryFilters(error_id=[id_2]))
    assert len(err) == 1
    assert err[0]["error_text"] == error_data_2["error_text"]

    # query by time
    err = storage_socket.serverinfo.query_error_log(ErrorLogQueryFilters(before=time_12))
    assert len(err) == 1
    assert err[0]["error_text"] == error_data_1["error_text"]

    err = storage_socket.serverinfo.query_error_log(ErrorLogQueryFilters(after=now_at_utc()))
    assert len(err) == 0

    err = storage_socket.serverinfo.query_error_log(ErrorLogQueryFilters(before=now_at_utc(), after=time_12))
    assert len(err) == 1

    err = storage_socket.serverinfo.query_error_log(ErrorLogQueryFilters(after=now_at_utc(), before=time_12))
    assert len(err) == 0

    # query by user
    err = storage_socket.serverinfo.query_error_log(ErrorLogQueryFilters(user=["read_user"]))
    assert len(err) == 1



"""Reaction spec hashing

Revision ID: a3c51b03bc19
Revises: 12d37bad13ba
Create Date: 2024-12-17 17:12:43.414568

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from qcfractal.alembic.migration_helpers.hashing import hash_dict_1

# revision identifiers, used by Alembic.
revision = "a3c51b03bc19"
down_revision = "12d37bad13ba"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "reaction_specification", sa.Column("protocols", postgresql.JSONB(astext_type=sa.Text()), nullable=True)
    )
    op.add_column("reaction_specification", sa.Column("specification_hash", sa.String(), nullable=True))

    # Generate hashes for existing specs
    conn = op.get_bind()
    res = conn.execute(sa.text(f"SELECT id, program, keywords FROM reaction_specification;"))
    all_spec = res.fetchall()
    for spec_id, program, keywords in all_spec:
        d = {
            "program": program,
            "keywords": keywords,
            "protocols": {},  # Default for now
        }
        h = hash_dict_1(d)
        op.execute(sa.text(f"""UPDATE reaction_specification SET specification_hash = '{h}' WHERE id = {spec_id};"""))
    op.execute(sa.text("UPDATE reaction_specification SET protocols = '{}'::JSONB"))

    op.alter_column("reaction_specification", "protocols", nullable=False)
    op.alter_column("reaction_specification", "specification_hash", nullable=False)
    op.drop_constraint("ux_reaction_specification_keys", "reaction_specification", type_="unique")
    op.drop_column("reaction_specification", "keywords_hash")
    op.create_unique_constraint(
        "ux_reaction_specification_keys",
        "reaction_specification",
        ["specification_hash", "singlepoint_specification_id", "optimization_specification_id"],
    )
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Downgrade not supported.")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from sqlalchemy import select, delete, func
from sqlalchemy.dialects.postgresql import insert

from qcfractal.components.dataset_db_models import BaseDatasetORM
from qcfractal.components.internal_jobs.db_models import InternalJobORM
from qcfractal.components.auth.db_models import UserORM
from qcfractal.components.project_db_models import (
    ProjectORM,
    ProjectRecordORM,
    ProjectDatasetORM,
    ProjectAttachmentORM,
    ProjectInternalJobORM,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.db_socket.helpers import get_general
from qcportal.all_inputs import AllInputTypes
from qcportal.all_results import AllResultTypes
from qcportal.exceptions import MissingDataError, UserReportableError, AlreadyExistsError
from qcportal.internal_jobs import InternalJobStatusEnum
from qcportal.metadata_models import InsertCountsMetadata
from qcportal.project_models import ProjectAttachmentType
from qcportal.record_models import PriorityEnum, RecordStatusEnum

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.components.internal_jobs.status import JobProgress
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Dict, Any, Optional, Sequence, Iterable, List
    from typing import Iterable, List, Dict, Any, Union, Tuple
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from sqlalchemy.orm.session import Session


class ProjectSocket:
    """
    Socket for projects
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

    def add(
        self,
        name: str,
        description: str,
        tagline: str,
        tags: List[str],
        default_compute_tag: str,
        default_compute_priority: PriorityEnum,
        extras: Dict[str, Any],
        owner_user: Optional[Union[int, str]],
        existing_ok: bool,
        *,
        session: Optional[Session] = None,
    ) -> int:
        """
        Create a new project in the database

        If a project already exists with the same name, an exception is raised if existing_ok is False

        Returns
        -------
        :
            ID of the new dataset
        """

        proj_orm = ProjectORM(
            name=name,
            description=description,
            tagline=tagline,
            tags=tags,
            default_compute_tag=default_compute_tag.lower(),
            default_compute_priority=default_compute_priority,
            extras=extras,
        )

        with self.root_socket.optional_session(session) as session:
            owner_user_id = self.root_socket.users.get_optional_user_id(owner_user)

            stmt = select(ProjectORM.id)
            stmt = stmt.where(ProjectORM.lname == name.lower())
            existing_id = session.execute(stmt).scalar_one_or_none()

            if existing_id is not None:
                if existing_ok:
                    return existing_id
                else:
                    raise AlreadyExistsError(f"Project with name='{name}' already exists")

            proj_orm.owner_user_id = owner_user_id

            session.add(proj_orm)
            session.commit()
            return proj_orm.id

    def get(
        self,
        project_id: int,
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Obtain a project with the specified ID

        Parameters
        ----------
        project_id
            ID of a project
        include
            Which fields of the result to return. Default is to return all fields.
        exclude
            Remove these fields from the return. Default is to return all fields.
        missing_ok
           If set to True, None will be returned if the project does not exist. Otherwise,
           an exception is raised.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Project information as a dictionary
        """

        with self.root_socket.optional_session(session, True) as session:
            try:
                return get_general(session, ProjectORM, ProjectORM.id, [project_id], include, exclude, missing_ok)[0]
            except MissingDataError:
                raise MissingDataError(f"Could not find project with id '{project_id}'")

    def lookup_id(
        self, project_name: str, missing_ok: bool = False, *, session: Optional[Session] = None
    ) -> Optional[int]:
        """
        Look up a project ID given its project type and name
        """

        stmt = select(ProjectORM.id)
        stmt = stmt.where(ProjectORM.lname == project_name.lower())

        with self.root_socket.optional_session(session, True) as session:
            ds_id = session.execute(stmt).scalar_one_or_none()

            if missing_ok is False and ds_id is None:
                raise MissingDataError(f"Could not find project with name '{project_name}'")
            return ds_id

    def list(self, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        """
        Get a list of projects in the database
        """

        record_count_cte = (
            select(ProjectRecordORM.project_id, func.count("*").label("record_count"))
            .group_by(ProjectRecordORM.project_id)
            .cte()
        )
        dataset_count_cte = (
            select(ProjectDatasetORM.project_id, func.count("*").label("dataset_count"))
            .group_by(ProjectDatasetORM.project_id)
            .cte()
        )

        with self.root_socket.optional_session(session, True) as session:
            stmt = select(
                ProjectORM.id,
                ProjectORM.name,
                ProjectORM.tagline,
                ProjectORM.tags,
                func.coalesce(record_count_cte.c.record_count, 0),
                func.coalesce(dataset_count_cte.c.dataset_count, 0),
                UserORM.username,
            )
            stmt = stmt.join(record_count_cte, ProjectORM.id == record_count_cte.c.project_id, isouter=True)
            stmt = stmt.join(dataset_count_cte, ProjectORM.id == dataset_count_cte.c.project_id, isouter=True)
            stmt = stmt.join(UserORM, ProjectORM.owner_user_id == UserORM.id, isouter=True)
            stmt = stmt.order_by(ProjectORM.id.asc())
            r = session.execute(stmt).all()

            return [
                {
                    "id": x[0],
                    "project_name": x[1],
                    "tagline": x[2],
                    "tags": x[3],
                    "record_count": x[4],
                    "dataset_count": x[5],
                    "owner_user": x[6],
                }
                for x in r
            ]

    def delete_project(
        self,
        project_id: int,
        delete_records: bool,
        delete_datasets: bool,
        delete_dataset_records: bool,
        *,
        session: Optional[Session] = None,
    ):
        """
        Deletes a project from the database

        Parameters
        ----------
        project_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:
            if delete_records:
                rec_ids = self.get_record_ids(project_id, session=session)
                self.unlink_records(project_id, rec_ids, delete_records=True, session=session)
            if delete_datasets:
                ds_ids = self.get_dataset_ids(project_id, session=session)
                self.unlink_datasets(
                    project_id,
                    ds_ids,
                    delete_datasets=True,
                    delete_dataset_records=delete_dataset_records,
                    session=session,
                )

            stmt = delete(ProjectORM)
            stmt = stmt.where(ProjectORM.id == project_id)
            session.execute(stmt)

    #################################
    # Status and General Info
    #################################
    def status(self, project_id: int, *, session: Optional[Session] = None) -> Dict[str, Dict[RecordStatusEnum, int]]:

        total_status = {}

        with self.root_socket.optional_session(session, True) as session:
            # Record status
            stmt = select(BaseRecordORM.status, func.count(BaseRecordORM.id))
            stmt = stmt.join(ProjectRecordORM, BaseRecordORM.id == ProjectRecordORM.record_id)
            stmt = stmt.where(ProjectRecordORM.project_id == project_id)
            stmt = stmt.group_by(BaseRecordORM.status)

            record_status = session.execute(stmt).all()
            total_status["records"] = {r[0]: r[1] for r in record_status}

            # Dataset status
            # do one at a time
            total_status["datasets"] = {}

            ds_ids = self.get_dataset_ids(project_id, session=session)
            for ds_id in ds_ids:
                ds_status = self.root_socket.datasets.overall_status(ds_id)
                for k, v in ds_status.items():
                    total_status["datasets"].setdefault(k, 0)
                    total_status["datasets"][k] += v

            return total_status

    #################################
    # Datasets
    #################################

    def assert_dataset_belongs(self, project_id: int, dataset_id: int, *, session: Optional[Session] = None):

        stmt = select(ProjectDatasetORM.dataset_id)
        stmt = stmt.where(ProjectDatasetORM.project_id == project_id)
        stmt = stmt.where(ProjectDatasetORM.dataset_id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            ds_id = session.execute(stmt).scalar_one_or_none()

            if ds_id is None:
                raise MissingDataError(f"Dataset {dataset_id} not found in project {project_id}")

    def dataset_name_exists(self, project_id: int, dataset_name: str, *, session: Optional[Session] = None):
        stmt = select(ProjectDatasetORM.dataset_id)
        stmt = stmt.join(BaseDatasetORM, ProjectDatasetORM.dataset_id == BaseDatasetORM.id)
        stmt = stmt.where(ProjectDatasetORM.project_id == project_id)
        stmt = stmt.where(BaseDatasetORM.lname == dataset_name.lower())

        with self.root_socket.optional_session(session, True) as session:
            ds_id = session.execute(stmt).scalar_one_or_none()
            return ds_id is not None

    def get_dataset_metadata(self, project_id: int, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        stmt = select(
            ProjectDatasetORM,
            BaseDatasetORM.dataset_type,
        )
        stmt = stmt.join(ProjectDatasetORM, BaseDatasetORM.id == ProjectDatasetORM.dataset_id)
        stmt = stmt.where(ProjectDatasetORM.project_id == project_id)

        with self.root_socket.optional_session(session, True) as session:
            meta_info = session.execute(stmt).all()
            return [
                {
                    "dataset_id": pd.dataset_id,
                    "dataset_type": ds_type,
                    "name": pd.name,
                    "description": pd.description,
                    "tagline": pd.tagline,
                    "tags": pd.tags,
                }
                for pd, ds_type in meta_info
            ]

    def get_dataset_ids(self, project_id: int, *, session: Optional[Session] = None) -> List[int]:
        stmt = select(ProjectDatasetORM.dataset_id).where(ProjectDatasetORM.project_id == project_id)

        with self.root_socket.optional_session(session, True) as session:
            rids = session.execute(stmt).scalars().all()
            return list(rids)

    def add_dataset(
        self,
        project_id: int,
        dataset_type: str,
        dataset_name: str,
        description: str,
        tagline: str,
        tags: List[str],
        provenance: Dict[str, Any],
        default_compute_tag: str,
        default_compute_priority: PriorityEnum,
        extras: Dict[str, Any],
        creator_user: Optional[Union[int, str]],
        existing_ok: bool,
        *,
        session: Optional[Session] = None,
    ) -> int:

        ds_socket = self.root_socket.datasets.get_socket(dataset_type)

        with self.root_socket.optional_session(session) as session:
            if self.dataset_name_exists(project_id, dataset_name, session=session):
                raise ValueError(f"Dataset '{dataset_name}' already exists in project {project_id}")

            # Note - name, description, tagline, and tags gets duplicated in places - in the dataset, and in the
            # link between the project and the dataset
            # This should be fixed at some point
            ds_id = ds_socket.add(
                name=dataset_name,
                description=description,
                tagline=tagline,
                tags=tags,
                provenance=provenance,
                default_compute_tag=default_compute_tag,
                default_compute_priority=default_compute_priority,
                extras=extras,
                creator_user=creator_user,
                existing_ok=existing_ok,
                session=session,
            )

            proj_ds_orm = ProjectDatasetORM(
                project_id=project_id,
                dataset_id=ds_id,
                name=dataset_name,
                description=description,
                tagline=tagline,
                tags=tags,
            )
            session.add(proj_ds_orm)
            return ds_id

    def get_dataset(
        self,
        project_id: int,
        dataset_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Any]:

        stmt = select(ProjectDatasetORM, BaseDatasetORM)
        stmt = stmt.join(ProjectDatasetORM, BaseDatasetORM.id == ProjectDatasetORM.dataset_id)
        stmt = stmt.where(ProjectDatasetORM.project_id == project_id)
        stmt = stmt.where(ProjectDatasetORM.dataset_id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            ds_info = session.execute(stmt).one_or_none()

            if ds_info is None:
                raise MissingDataError(f"Dataset {dataset_id} not found in project {project_id}")

            # Use the name, description, etc from the project, not from the base dataset
            project_ds, base_ds = ds_info

            ds_dict = base_ds.model_dict()
            ds_dict.update(
                name=project_ds.name,
                description=project_ds.description,
                tagline=project_ds.tagline,
                tags=project_ds.tags,
            )

            return ds_dict

    def link_dataset(
        self,
        project_id: int,
        dataset_id: int,
        name: Optional[str],
        description: Optional[str],
        tagline: Optional[str],
        tags: Optional[List[str]],
        *,
        session: Optional[Session] = None,
    ):
        with self.root_socket.optional_session(session, True) as session:
            ds = self.root_socket.datasets.get(
                dataset_id, include={"name", "description", "tagline", "tags"}, missing_ok=False, session=session
            )

            stmt = (
                insert(ProjectDatasetORM)
                .values(
                    project_id=project_id,
                    dataset_id=dataset_id,
                    name=ds["name"] if name is None else name,
                    description=ds["description"] if description is None else description,
                    tagline=ds["tagline"] if tagline is None else tagline,
                    tags=ds["tags"] if tags is None else tags,
                )
                .on_conflict_do_nothing()
                .returning(ProjectDatasetORM.dataset_id)
            )

            r = session.execute(stmt).scalar_one_or_none()

            if r is None:
                raise ValueError(f"Dataset {dataset_id} already linked to project {project_id}")

    def unlink_datasets(
        self,
        project_id: int,
        dataset_ids: Iterable[int],
        delete_datasets: bool,
        delete_dataset_records: bool,
        *,
        session: Optional[Session] = None,
    ):

        stmt = delete(ProjectDatasetORM)
        stmt = stmt.where(ProjectDatasetORM.project_id == project_id)
        stmt = stmt.where(ProjectDatasetORM.dataset_id.in_(dataset_ids))
        stmt = stmt.returning(ProjectDatasetORM.dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            ds_ids = session.execute(stmt).scalars().all()

            # Use ds_ids so we only delete datasets that were removed from this dataset
            if delete_datasets:
                for ds_id in ds_ids:
                    self.root_socket.datasets.delete(ds_id, delete_dataset_records, session=session)

    #################################
    # Records
    #################################

    def assert_record_belongs(self, project_id: int, record_id: int, *, session: Optional[Session] = None):

        stmt = select(ProjectRecordORM.record_id)
        stmt = stmt.where(ProjectRecordORM.project_id == project_id)
        stmt = stmt.where(ProjectRecordORM.record_id == record_id)

        with self.root_socket.optional_session(session, True) as session:
            r_id = session.execute(stmt).scalar_one_or_none()

            if r_id is None:
                raise MissingDataError(f"Record {record_id} not found in project {project_id}")

    def record_name_exists(self, project_id: int, record_name: str, *, session: Optional[Session] = None):
        stmt = select(ProjectRecordORM.record_id)
        stmt = stmt.where(ProjectRecordORM.project_id == project_id)
        stmt = stmt.where(ProjectRecordORM.lname == record_name.lower())

        with self.root_socket.optional_session(session, True) as session:
            r_id = session.execute(stmt).scalar_one_or_none()
            return r_id is not None

    def get_record_metadata(self, project_id: int, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        stmt = select(ProjectRecordORM, BaseRecordORM.record_type, BaseRecordORM.status)
        stmt = stmt.join(BaseRecordORM, BaseRecordORM.id == ProjectRecordORM.record_id)
        stmt = stmt.where(ProjectRecordORM.project_id == project_id)

        with self.root_socket.optional_session(session, True) as session:
            meta_info = session.execute(stmt).all()
            return [
                {"record_type": record_type, "status": status, **pr.model_dict()}
                for pr, record_type, status in meta_info
            ]

    def get_record_ids(self, project_id: int, *, session: Optional[Session] = None) -> List[int]:
        stmt = select(ProjectRecordORM.record_id).where(ProjectRecordORM.project_id == project_id)

        with self.root_socket.optional_session(session, True) as session:
            rids = session.execute(stmt).scalars().all()
            return list(rids)

    def add_record(
        self,
        project_id: int,
        name: str,
        description: str,
        tags: List[str],
        record_input: AllInputTypes,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertCountsMetadata, int]:

        with self.root_socket.optional_session(session) as session:
            if self.record_name_exists(project_id, name, session=session):
                raise ValueError(f"Record '{name}' already exists in project {project_id}")

            meta, record_id = self.root_socket.records.add_from_input(
                record_input,
                compute_tag=compute_tag,
                compute_priority=compute_priority,
                creator_user=creator_user,
                find_existing=find_existing,
                session=session,
            )

            if meta.success:
                proj_rec_orm = ProjectRecordORM(
                    project_id=project_id,
                    record_id=record_id,
                    name=name,
                    description=description,
                    tags=tags,
                )
                session.add(proj_rec_orm)

            return meta, record_id

    def import_record(
        self,
        project_id: int,
        name: str,
        description: str,
        tags: List[str],
        record_data: AllResultTypes,
        creator_user: Optional[Union[int, str]],
        *,
        session: Optional[Session] = None,
    ) -> int:

        with self.root_socket.optional_session(session) as session:
            record_ids = self.root_socket.records.insert_full_record(session, [record_data], creator_user)
            record_id = record_ids[0]

            proj_rec_orm = ProjectRecordORM(
                project_id=project_id,
                record_id=record_id,
                name=name,
                description=description,
                tags=tags,
            )

            session.add(proj_rec_orm)
            return record_id

    def get_record(
        self,
        project_id: int,
        record_id: int,
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Any]:

        stmt = select(ProjectRecordORM, BaseRecordORM.record_type)
        stmt = stmt.join(ProjectRecordORM, BaseRecordORM.id == ProjectRecordORM.record_id)
        stmt = stmt.where(ProjectRecordORM.project_id == project_id)
        stmt = stmt.where(ProjectRecordORM.record_id == record_id)

        with self.root_socket.optional_session(session, True) as session:
            project_rec = session.execute(stmt).one_or_none()

            if project_rec is None:
                raise MissingDataError(f"Record {record_id} not found in project {project_id}")

            pr, record_type = project_rec
            record_socket = self.root_socket.records.get_socket(record_type)
            rec_dict = record_socket.get(
                [record_id], include=include, exclude=exclude, missing_ok=False, session=session
            )[0]
            rec_dict.update(name=pr.name, description=pr.description, tags=pr.tags)
            return rec_dict

    def link_record(
        self,
        project_id: int,
        record_id: int,
        name: str,
        description: str,
        tags: List[str],
        *,
        session: Optional[Session] = None,
    ):
        with self.root_socket.optional_session(session, True) as session:
            stmt = (
                insert(ProjectRecordORM)
                .values(
                    project_id=project_id,
                    record_id=record_id,
                    name=name,
                    description=description,
                    tags=tags,
                )
                .on_conflict_do_nothing()
                .returning(ProjectRecordORM.record_id)
            )

            r = session.execute(stmt).scalar_one_or_none()

            if r is None:
                raise ValueError(f"Record {record_id} already linked to project {project_id}")

    def unlink_records(
        self, project_id: int, record_ids: Iterable[int], delete_records: bool, *, session: Optional[Session] = None
    ):

        stmt = delete(ProjectRecordORM)
        stmt = stmt.where(ProjectRecordORM.project_id == project_id)
        stmt = stmt.where(ProjectRecordORM.record_id.in_(record_ids))
        stmt = stmt.returning(ProjectRecordORM.record_id)

        with self.root_socket.optional_session(session, True) as session:
            r_ids = session.execute(stmt).scalars().all()

            # Use r_ids so we only delete records that were removed from this dataset
            if delete_records:
                self.root_socket.records.delete(r_ids, False, session=session)

    #################################
    # Internal Jobs
    #################################

    def get_internal_job(
        self,
        project_id: int,
        job_id: int,
        *,
        session: Optional[Session] = None,
    ):
        stmt = select(InternalJobORM)
        stmt = stmt.join(ProjectInternalJobORM, ProjectInternalJobORM.internal_job_id == InternalJobORM.id)
        stmt = stmt.where(ProjectInternalJobORM.project_id == project_id)
        stmt = stmt.where(ProjectInternalJobORM.internal_job_id == job_id)

        with self.root_socket.optional_session(session, True) as session:
            ij_orm = session.execute(stmt).scalar_one_or_none()
            if ij_orm is None:
                raise MissingDataError(f"Job id {job_id} not found in project {project_id}")
            return ij_orm.model_dict()

    def list_internal_jobs(
        self,
        project_id: int,
        status: Optional[Iterable[InternalJobStatusEnum]] = None,
        *,
        session: Optional[Session] = None,
    ):
        stmt = select(InternalJobORM)
        stmt = stmt.join(ProjectInternalJobORM, ProjectInternalJobORM.internal_job_id == InternalJobORM.id)
        stmt = stmt.where(ProjectInternalJobORM.project_id == project_id)

        if status is not None:
            stmt = stmt.where(InternalJobORM.status.in_(status))

        with self.root_socket.optional_session(session, True) as session:
            ij_orm = session.execute(stmt).scalars().all()
            return [i.model_dict() for i in ij_orm]

    #################################
    # Attachments
    #################################
    def get_attachments(self, project_id: int, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        """
        Get the attachments for a project
        """

        stmt = select(ProjectAttachmentORM)
        stmt = stmt.where(ProjectAttachmentORM.project_id == project_id)

        with self.root_socket.optional_session(session, True) as session:
            att = session.execute(stmt).scalars().all()
            return [x.model_dict() for x in att]

    def delete_attachment(self, project_id: int, file_id: int, *, session: Optional[Session] = None):
        stmt = select(ProjectAttachmentORM)
        stmt = stmt.where(ProjectAttachmentORM.project_id == project_id)
        stmt = stmt.where(ProjectAttachmentORM.id == file_id)
        stmt = stmt.with_for_update()

        with self.root_socket.optional_session(session) as session:
            att = session.execute(stmt).scalar_one_or_none()
            if att is None:
                raise MissingDataError(f"Attachment with file id {file_id} not found in project {project_id}")

            return self.root_socket.external_files.delete(file_id, session=session)

    def attach_file(
        self,
        project_id: int,
        attachment_type: ProjectAttachmentType,
        file_path: str,
        file_name: str,
        description: Optional[str],
        *,
        job_progress: Optional[JobProgress] = None,
        session: Optional[Session] = None,
    ) -> int:
        """
        Attach a file to a project

        This function uploads the specified file and associates it with the project by creating
        a corresponding project attachment record. This operation requires S3 storage to be enabled.

        Parameters
        ----------
        project_id
            The ID of the project to which the file will be attached.
        attachment_type
            The type of attachment that categorizes the file being added.
        file_path
            The local file system path to the file that needs to be uploaded.
        file_name
            The name of the file to be used in the attachment record. This is the filename that is
            recommended to the user by default.
        description
            An optional description of the file
        job_progress
            Object used to track progress if this function is being run in a background job
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Raises
        ------
        UserReportableError
            Raised if S3 storage is not enabled
        """

        if not self.root_socket.qcf_config.s3.enabled:
            raise UserReportableError("S3 storage is not enabled. Can not attach file to a project")

        self._logger.info(f"Uploading/Attaching project-related file: {file_path}")
        with self.root_socket.optional_session(session) as session:
            ef = ProjectAttachmentORM(
                project_id=project_id,
                attachment_type=attachment_type,
                file_name=file_name,
                description=description,
            )

            file_id = self.root_socket.external_files.add_file(
                file_path, ef, session=session, job_progress=job_progress
            )

            self._logger.info(f"Project attachment {file_path} successfully uploaded to S3. ID is {file_id}")
            return file_id



from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING

from sqlalchemy import select, literal, text, insert

from qcfractal.components.singlepoint.record_db_models import SinglepointRecordORM
from qcportal.exceptions import InvalidArgumentsError, MissingDataError
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.record_models import PriorityEnum
from qcportal.singlepoint import SinglepointDatasetNewEntry, QCSpecification
from .dataset_db_models import (
    SinglepointDatasetORM,
    SinglepointDatasetSpecificationORM,
    SinglepointDatasetEntryORM,
    SinglepointDatasetRecordItemORM,
)
from ..base_dataset_socket import BaseDatasetSocket

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Optional, Sequence, Iterable, Tuple


class SinglepointDatasetSocket(BaseDatasetSocket):
    # Used by the base class
    dataset_orm = SinglepointDatasetORM
    specification_orm = SinglepointDatasetSpecificationORM
    entry_orm = SinglepointDatasetEntryORM
    record_item_orm = SinglepointDatasetRecordItemORM
    record_orm = SinglepointRecordORM

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseDatasetSocket.__init__(
            self,
            root_socket,
        )

        self._logger = logging.getLogger(__name__)

    def _add_specification(
        self, session: Session, specification: QCSpecification
    ) -> Tuple[InsertMetadata, Optional[int]]:
        return self.root_socket.records.singlepoint.add_specification(specification, session=session)

    def _create_entries(self, session: Session, dataset_id: int, new_entries: Sequence[SinglepointDatasetNewEntry]):
        molecules = [x.molecule for x in new_entries]
        meta, mol_ids = self.root_socket.molecules.add_mixed(molecules, session=session)

        all_entries = []
        for entry, molecule_id in zip(new_entries, mol_ids):
            new_ent = SinglepointDatasetEntryORM(
                dataset_id=dataset_id,
                name=entry.name,
                comment=entry.comment,
                molecule_id=molecule_id,
                additional_keywords=entry.additional_keywords,
                attributes=entry.attributes,
                local_results=entry.local_results,
            )

            all_entries.append(new_ent)

        return all_entries

    def _submit(
        self,
        session: Session,
        dataset_id: int,
        entry_orm: Iterable[SinglepointDatasetEntryORM],
        spec_orm: Iterable[SinglepointDatasetSpecificationORM],
        existing_records: Iterable[Tuple[str, str]],
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
    ) -> InsertCountsMetadata:

        n_inserted = 0
        n_existing = 0

        # Weed out any with additional keywords
        special_entries = [x for x in entry_orm if x.additional_keywords]
        normal_entries = [x for x in entry_orm if not x.additional_keywords]

        # Normal entries - just let it rip
        for spec in spec_orm:
            new_normal_entries = [x for x in normal_entries if (x.name, spec.name) not in existing_records]
            molecule_ids = [x.molecule_id for x in new_normal_entries]

            meta, sp_ids = self.root_socket.records.singlepoint.add_internal(
                molecule_ids=molecule_ids,
                qc_spec_id=spec.specification_id,
                compute_tag=compute_tag,
                compute_priority=compute_priority,
                creator_user_id=creator_user_id,
                find_existing=find_existing,
                session=session,
            )

            for entry, oid in zip(new_normal_entries, sp_ids):
                rec = SinglepointDatasetRecordItemORM(
                    dataset_id=dataset_id, entry_name=entry.name, specification_name=spec.name, record_id=oid
                )
                session.add(rec)

            n_inserted += meta.n_inserted
            n_existing += meta.n_existing

        # Now the ones with additional keywords
        for spec in spec_orm:
            spec_obj = spec.specification.to_model(QCSpecification)
            spec_input_dict = spec_obj.dict()

            for entry in special_entries:
                if (entry.name, spec.name) in existing_records:
                    continue

                new_spec = copy.deepcopy(spec_input_dict)
                new_spec["keywords"].update(entry.additional_keywords)

                meta, sp_ids = self.root_socket.records.singlepoint.add(
                    molecules=[entry.molecule_id],
                    qc_spec=QCSpecification(**new_spec),
                    compute_tag=compute_tag,
                    compute_priority=compute_priority,
                    creator_user=creator_user_id,
                    find_existing=find_existing,
                    session=session,
                )

                if meta.success:
                    rec = SinglepointDatasetRecordItemORM(
                        dataset_id=dataset_id,
                        entry_name=entry.name,
                        specification_name=spec.name,
                        record_id=sp_ids[0],
                    )
                    session.add(rec)

                n_inserted += meta.n_inserted
                n_existing += meta.n_existing

        return InsertCountsMetadata(n_inserted=n_inserted, n_existing=n_existing)

    def _copy_entries(
        self,
        session: Session,
        source_dataset_id: int,
        destination_dataset_id: int,
        entry_names: Optional[Iterable[str]] = None,
    ):

        select_stmt = select(
            literal(destination_dataset_id),
            self.entry_orm.name,
            self.entry_orm.comment,
            self.entry_orm.molecule_id,
            self.entry_orm.additional_keywords,
            self.entry_orm.attributes,
            self.entry_orm.local_results,
        )

        select_stmt = select_stmt.where(self.entry_orm.dataset_id == source_dataset_id)

        if entry_names is not None:
            select_stmt = select_stmt.where(self.entry_orm.name.in_(entry_names))

        stmt = insert(self.entry_orm)
        stmt = stmt.from_select(
            [
                self.entry_orm.dataset_id,
                self.entry_orm.name,
                self.entry_orm.comment,
                self.entry_orm.molecule_id,
                self.entry_orm.additional_keywords,
                self.entry_orm.attributes,
                self.entry_orm.local_results,
            ],
            select_stmt,
        )

        session.execute(stmt)

    def add_entries_from_ds(
        self,
        dataset_id: int,
        from_dataset_id: Optional[int],
        from_dataset_type: Optional[str],
        from_dataset_name: Optional[str],
        from_specification_name: Optional[str],
        *,
        session: Optional[Session] = None,
    ) -> InsertCountsMetadata:
        """
        Adds entries from another dataset into this Singlepoint dataset

        Either the from_dataset_id or both the from_dataset_type and from_dataset_name
        must be provided.

        If copying from an optimization dataset, then
        the from_specification_name must be provided.

        Parameters
        ----------
        dataset_id
            ID of the dataset to add entries to
        from_dataset_id
            ID of the dataset to add entries from
        from_dataset_type
            Type of the dataset to add entries from
        from_dataset_name
            Name of the dataset to add entries from
        from_specification_name
            Specification of the source dataset to use for molecules

        Returns
        -------
        :
            Metadata about the added entries
        """

        if dataset_id is None and (from_dataset_type is None or from_dataset_name is None):
            raise InvalidArgumentsError("dataset_id or both from_dataset_id and from_dataset_name must be provided")

        with self.root_socket.optional_session(session) as session:
            # Make sure dataset exists
            stmt = "SELECT (id) FROM base_dataset WHERE id = :dataset_id"
            r = session.execute(text(stmt), {"dataset_id": dataset_id}).scalar_one_or_none()
            if r is None:
                raise MissingDataError(f"Cannot find dataset with id={dataset_id}")

            # No matter what was passed in, we need the dataset type and id to copy from
            if from_dataset_id is not None:
                stmt = "SELECT (dataset_type) FROM base_dataset WHERE id = :from_dataset_id"
                ds_type = session.execute(text(stmt), {"from_dataset_id": from_dataset_id}).scalar_one_or_none()

                if ds_type is None:
                    raise MissingDataError(f"Cannot find dataset with id={from_dataset_id}")

                if from_dataset_type is not None and from_dataset_type.lower() != ds_type.lower():
                    raise InvalidArgumentsError(
                        f"Dataset id and type specified, but the type of the dataset id={from_dataset_id} is {ds_type}, not {from_dataset_type}"
                    )
                from_dataset_type = ds_type
            else:
                assert from_dataset_type is not None and from_dataset_name is not None
                stmt = "SELECT (id) FROM base_dataset WHERE dataset_type = :from_dataset_type and lname = :from_dataset_name"
                from_dataset_id = session.execute(
                    text(stmt), {"from_dataset_type": from_dataset_type, "from_dataset_name": from_dataset_name.lower()}
                ).scalar_one_or_none()

                if from_dataset_id is None:
                    raise MissingDataError(
                        f"Cannot find dataset with type={from_dataset_type} and name={from_dataset_name}"
                    )

            if from_dataset_type == "singlepoint":
                stmt = """
                    INSERT INTO singlepoint_dataset_entry (dataset_id, name, comment, molecule_id, attributes, additional_keywords)
                    SELECT :dataset_id, sde.name, sde.comment, sde.molecule_id, sde.attributes, sde.additional_keywords
                    FROM singlepoint_dataset_entry sde
                    WHERE sde.dataset_id = :from_dataset_id
                    ON CONFLICT (dataset_id, name) DO NOTHING
                    RETURNING 1
                """

                r = session.execute(
                    text(stmt),
                    {
                        "dataset_id": dataset_id,
                        "from_dataset_id": from_dataset_id,
                    },
                )

                meta = InsertCountsMetadata(n_inserted=r.rowcount, n_existing=0)

            elif from_dataset_type == "optimization":

                if from_specification_name is None:
                    raise InvalidArgumentsError(
                        "from_specification_name must be provided when adding entries from an optimization dataset"
                    )

                stmt = """
                    INSERT INTO singlepoint_dataset_entry (dataset_id, name, comment, molecule_id, attributes, additional_keywords)
                    SELECT :dataset_id, ode.name, ode.comment, opr.final_molecule_id, ode.attributes, '{}'::jsonb
                    FROM optimization_dataset_entry ode
                    INNER JOIN optimization_dataset_record odr ON ode.dataset_id = odr.dataset_id AND ode.name = odr.entry_name
                    INNER JOIN optimization_record opr ON odr.record_id = opr.id
                    INNER JOIN base_record br ON opr.id = br.id
                    WHERE ode.dataset_id = :optimization_dataset_id
                    AND odr.specification_name = :specification_name
                    AND br.status = 'complete'
                    AND opr.final_molecule_id IS NOT NULL
                    ON CONFLICT (dataset_id, name) DO NOTHING
                    RETURNING 1
                """

                r = session.execute(
                    text(stmt),
                    {
                        "dataset_id": dataset_id,
                        "optimization_dataset_id": from_dataset_id,
                        "specification_name": from_specification_name,
                    },
                )

                meta = InsertCountsMetadata(n_inserted=r.rowcount, n_existing=0)
            else:
                raise InvalidArgumentsError(
                    f"Unable to handle adding singlepoint dataset entries from dataset of type {from_dataset_type}"
                )

        return meta



from typing import Optional

from flask import current_app, g

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.base_models import ProjURLParameters, CommonBulkGetBody
from qcportal.exceptions import LimitExceededError
from qcportal.record_models import (
    RecordModifyBody,
    RecordQueryFilters,
    RecordDeleteBody,
    RecordRevertBody,
)


#################################################################
# Base record route
# A few things can be done directly through /records (rather than
# a /records/<record_type>/ route
#################################################################


@api_v1.route("/records/query", methods=["POST"])
@wrap_global_route("records", "read")
def query_records_v1(body_data: RecordQueryFilters):
    return storage_socket.records.query(body_data)


@api_v1.route("/records/revert", methods=["POST"])
@wrap_global_route("records", "modify")
def revert_records_v1(body_data: RecordRevertBody):
    return storage_socket.records.revert_generic(body_data.record_ids, body_data.revert_status)


@api_v1.route("/records", methods=["PATCH"])
@wrap_global_route("records", "modify")
def modify_records_v1(body_data: RecordModifyBody):
    return storage_socket.records.modify_generic(
        body_data.record_ids,
        g.user_id,
        body_data.status,
        body_data.compute_priority,
        body_data.compute_tag,
        body_data.comment,
    )


@api_v1.route("/records/<int:record_id>", methods=["DELETE"])
@wrap_global_route("records", "delete")
def delete_records_v1(record_id: int):
    return storage_socket.records.delete([record_id], soft_delete=True, delete_children=True)


@api_v1.route("/records/bulkDelete", methods=["POST"])
@wrap_global_route("records", "delete")
def bulk_delete_records_v1(body_data: RecordDeleteBody):
    return storage_socket.records.delete(
        body_data.record_ids, soft_delete=body_data.soft_delete, delete_children=body_data.delete_children
    )


@api_v1.route("/records/<int:record_id>/waiting_reason", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_waiting_reason_v1(record_id: int):
    return storage_socket.records.get_waiting_reason(record_id)


#################################################################
# Routes for individual record types
# These can also be accessed through /records
#################################################################


@api_v1.route("/records/<string:record_type>/<int:record_id>", methods=["GET"])
@api_v1.route("/records/<int:record_id>", methods=["GET"])
@wrap_global_route("records", "read")
def get_records_v1(record_id: int, url_params: ProjURLParameters, record_type: Optional[str] = None):
    if record_type is None:
        return storage_socket.records.get([record_id], url_params.include, url_params.exclude)[0]
    else:
        record_socket = storage_socket.records.get_socket(record_type)
        return record_socket.get([record_id], url_params.include, url_params.exclude)[0]


@api_v1.route("/records/<string:record_type>/bulkGet", methods=["POST"])
@api_v1.route("/records/bulkGet", methods=["POST"])
@wrap_global_route("records", "read")
def bulk_get_records_v1(body_data: CommonBulkGetBody, record_type: Optional[str] = None):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_records
    if len(body_data.ids) > limit:
        raise LimitExceededError(f"Cannot get {len(body_data.ids)} records - limit is {limit}")

    # Getting is handled a little differently. If no type specified, use the more generic version
    # in the upper-level record socket
    if record_type is None:
        return storage_socket.records.get(body_data.ids, body_data.include, body_data.exclude, body_data.missing_ok)
    else:
        record_socket = storage_socket.records.get_socket(record_type)
        return record_socket.get(body_data.ids, body_data.include, body_data.exclude, body_data.missing_ok)


@api_v1.route("/records/<string:record_type>/<int:record_id>", methods=["GET"])
@api_v1.route("/records/<int:record_id>", methods=["GET"])
@wrap_global_route("records", "read")
def get_general_records_v1(record_id: int, url_params: ProjURLParameters, record_type: Optional[str] = None):
    # Getting is handled a little differently. If no type specified, use the more generic version
    # in the upper-level record socket
    if record_type is None:
        records = storage_socket.records.get([record_id], url_params.include, url_params.exclude)
    else:
        record_socket = storage_socket.records.get_socket(record_type)
        records = record_socket.get([record_id], url_params.include, url_params.exclude)

    return records[0]


#################################################################
# Functions for getting individual record properties
# These are only accessible through routes for the individual
# record types
# But we can still handle them generically
#################################################################


@api_v1.route("/records/<string:record_type>/<int:record_id>/comments", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_comments_v1(record_id: int, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_comments(record_id)


@api_v1.route("/records/<string:record_type>/<int:record_id>/task", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_task_v1(record_id: int, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_task(record_id)


@api_v1.route("/records/<string:record_type>/<int:record_id>/service", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_service_v1(record_id: int, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_service(record_id)


@api_v1.route("/records/<string:record_type>/<int:record_id>/compute_history", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_history_v1(record_id: int, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_all_compute_history(record_id)


@api_v1.route("/records/<string:record_type>/<int:record_id>/compute_history/<int:history_id>", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_history_single_v1(record_id: int, history_id: int, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_single_compute_history(record_id, history_id)


@api_v1.route("/records/<string:record_type>/<int:record_id>/compute_history/<int:history_id>/outputs", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_outputs_v1(record_id: int, history_id: int, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_all_output_metadata(record_id, history_id)


@api_v1.route(
    "/records/<string:record_type>/<int:record_id>/compute_history/<int:history_id>/outputs/<string:output_type>",
    methods=["GET"],
)
@wrap_global_route("records", "read")
def get_record_outputs_single_v1(record_id: int, history_id: int, output_type: str, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_single_output_metadata(record_id, history_id, output_type)


@api_v1.route(
    "/records/<string:record_type>/<int:record_id>/compute_history/<int:history_id>/outputs/<string:output_type>/data",
    methods=["GET"],
)
@wrap_global_route("records", "read")
def get_record_outputs_data_v1(record_id: int, history_id: int, output_type: str, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_single_output_rawdata(record_id, history_id, output_type)


@api_v1.route(
    "/records/<string:record_type>/<int:record_id>/compute_history/<int:history_id>/outputs/<string:output_type>/uncompressed_data",
    methods=["GET"],
)
@wrap_global_route("records", "read")
def get_record_outputs_uncompressed_data_v1(record_id: int, history_id: int, output_type: str, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_single_output_uncompressed(record_id, history_id, output_type)


@api_v1.route("/records/<string:record_type>/<int:record_id>/native_files", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_native_files_v1(record_id: int, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_all_native_files_metadata(record_id)


@api_v1.route("/records/<string:record_type>/<int:record_id>/native_files/<string:name>", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_native_file_single_v1(record_id: int, name: str, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_single_native_file_metadata(record_id, name)


@api_v1.route("/records/<string:record_type>/<int:record_id>/native_files/<string:name>/data", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_native_file_data_v1(record_id: int, name: str, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_single_native_file_rawdata(record_id, name)


@api_v1.route(
    "/records/<string:record_type>/<int:record_id>/native_files/<string:name>/uncompressed_data", methods=["GET"]
)
@wrap_global_route("records", "read")
def get_record_native_file_uncompressed_data_v1(record_id: int, name: str, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_single_native_file_uncompressed(record_id, name)


@api_v1.route("/records/<string:record_type>/<int:record_id>/children_status", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_children_status_v1(record_id: int, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_children_status(record_id)


@api_v1.route("/records/<string:record_type>/<int:record_id>/children_errors", methods=["GET"])
@wrap_global_route("records", "read")
def get_record_children_errors_v1(record_id: int, record_type: str):
    record_socket = storage_socket.records.get_socket(record_type)
    return record_socket.get_children_errors(record_id)



from __future__ import annotations

import logging
from typing import List, Dict, Tuple, Optional, Iterable, Sequence, Any, Union, TYPE_CHECKING

import tabulate
from sqlalchemy import select, func
from sqlalchemy.dialects.postgresql import array_agg, aggregate_order_by
from sqlalchemy.orm import defer, undefer, joinedload, lazyload, selectinload

from qcfractal import __version__ as qcfractal_version
from qcfractal.components.optimization.record_db_models import OptimizationSpecificationORM
from qcfractal.components.services.db_models import ServiceQueueORM, ServiceDependencyORM
from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM
from qcfractal.db_socket.helpers import insert_general
from qcportal.exceptions import MissingDataError
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.molecules import Molecule
from qcportal.reaction import (
    ReactionSpecification,
    ReactionQueryFilters,
    ReactionInput,
    ReactionMultiInput,
    ReactionRecord,
)
from qcportal.record_models import PriorityEnum, RecordStatusEnum, OutputTypeEnum
from qcportal.utils import hash_dict, is_included
from .record_db_models import ReactionComponentORM, ReactionSpecificationORM, ReactionRecordORM
from ..base_record_socket import BaseRecordSocket
from ..record_utils import append_output

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket

# Meaningless, but unique to reaction
reaction_insert_lock_id = 14400


class ReactionRecordSocket(BaseRecordSocket):
    """
    Socket for handling reaction computations
    """

    # Used by the base class
    record_orm = ReactionRecordORM
    record_input_type = ReactionInput
    record_multi_input_type = ReactionMultiInput

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseRecordSocket.__init__(self, root_socket)
        self._logger = logging.getLogger(__name__)

    def initialize_service(self, session: Session, service_orm: ServiceQueueORM) -> None:
        rxn_orm: ReactionRecordORM = service_orm.record

        output = "\n\nCreated reaction. Molecules:\n\n"

        output += "-" * 80 + "\nManybody Keywords:\n\n"
        spec: ReactionSpecification = rxn_orm.specification.to_model(ReactionSpecification)
        table_rows = sorted(spec.keywords.dict().items())
        output += tabulate.tabulate(table_rows, headers=["keyword", "value"])

        if spec.singlepoint_specification:
            output += "\n\n" + "-" * 80 + "\nQC Specification:\n\n"
            table_rows = sorted(spec.singlepoint_specification.dict().items())
            output += tabulate.tabulate(table_rows, headers=["keyword", "value"])
            output += "\n\n"

        if spec.optimization_specification:
            output += "\n\n" + "-" * 80 + "\nOptimization Specification:\n\n"
            table_rows = sorted(spec.optimization_specification.dict().items())
            output += tabulate.tabulate(table_rows, headers=["keyword", "value"])
            output += "\n\n"

        output += "\n\n" + "-" * 80 + "\nReaction Stoichiometry:\n\n"
        table_rows = [
            (f"{m.coefficient:.8f}", m.molecule.identifiers["molecular_formula"], m.molecule_id)
            for m in rxn_orm.components
        ]
        output += tabulate.tabulate(table_rows, headers=["coefficient", "molecule", "molecule id"])
        output += "\n\n"

        append_output(session, rxn_orm, OutputTypeEnum.stdout, output)

        # Reactions are simple and don't require a service state

    def iterate_service(
        self,
        session: Session,
        service_orm: ServiceQueueORM,
    ):
        rxn_orm: ReactionRecordORM = service_orm.record

        spec: ReactionSpecification = rxn_orm.specification.to_model(ReactionSpecification)
        has_singlepoint = bool(spec.singlepoint_specification)
        has_optimization = bool(spec.optimization_specification)

        # Always update with the current provenance
        rxn_orm.compute_history[-1].provenance = {
            "creator": "qcfractal",
            "version": qcfractal_version,
            "routine": "qcfractal.services.reaction",
        }

        # Component by molecule id
        component_map = {x.molecule_id: x for x in rxn_orm.components}

        # Molecules we need to do an optimization on
        required_opt_mols = {x.molecule_id for x in rxn_orm.components if has_optimization}

        # Molecules we need to do a singlepoint calculation on
        # (this is the molecule of the reaction, so it may actually be the INITIAL molecule
        # of an optimization)
        required_sp_mols = {x.molecule_id for x in rxn_orm.components if has_singlepoint}

        # What was already completed and/or submitted
        sub_opt_mols = {x.molecule_id for x in rxn_orm.components if x.optimization_id is not None}
        sub_sp_mols = {x.molecule_id for x in rxn_orm.components if x.singlepoint_id is not None}

        # What we need to compute
        opt_mols_to_compute = required_opt_mols - sub_opt_mols
        sp_mols_to_compute = required_sp_mols - sub_sp_mols

        # Singlepoint calculations must wait for optimizations
        sp_mols_to_compute -= opt_mols_to_compute

        # Convert to well-ordered lists
        opt_mols_to_compute = list(opt_mols_to_compute)
        sp_mols_to_compute = list(sp_mols_to_compute)

        service_orm.dependencies = []
        output = ""

        if opt_mols_to_compute:
            opt_spec_id = rxn_orm.specification.optimization_specification_id

            meta, opt_ids = self.root_socket.records.optimization.add_internal(
                opt_mols_to_compute,
                opt_spec_id,
                service_orm.compute_tag,
                service_orm.compute_priority,
                rxn_orm.creator_user_id,
                service_orm.find_existing,
                session=session,
            )

            for mol_id, opt_id in zip(opt_mols_to_compute, opt_ids):
                component = component_map[mol_id]

                svc_dep = ServiceDependencyORM(record_id=opt_id, extras={})

                assert component.singlepoint_id is None
                component.optimization_id = opt_id

                service_orm.dependencies.append(svc_dep)

            output = "\n\nSubmitted optimization calculations:\n"
            output += tabulate.tabulate(zip(opt_mols_to_compute, opt_ids), headers=["molecule id", "optimization id"])

        if sp_mols_to_compute:
            # If an optimization was specified, we need to get the final molecule from that
            if has_optimization:
                real_mols_to_compute = [
                    x.optimization_record.final_molecule_id
                    for x in rxn_orm.components
                    if x.molecule_id in sp_mols_to_compute
                ]
            else:
                real_mols_to_compute = sp_mols_to_compute

            qc_spec_id = rxn_orm.specification.singlepoint_specification_id
            meta, sp_ids = self.root_socket.records.singlepoint.add_internal(
                real_mols_to_compute,
                qc_spec_id,
                service_orm.compute_tag,
                service_orm.compute_priority,
                rxn_orm.creator_user_id,
                service_orm.find_existing,
                session=session,
            )

            # Note the mapping back to the original molecule id (not the optimized one)
            for mol_id, sp_id in zip(sp_mols_to_compute, sp_ids):
                component = component_map[mol_id]

                svc_dep = ServiceDependencyORM(record_id=sp_id, extras={})

                assert component.singlepoint_id is None
                component.singlepoint_id = sp_id

                service_orm.dependencies.append(svc_dep)

            output = "\n\nSubmitted singlepoint calculations:\n"
            output += tabulate.tabulate(zip(sp_mols_to_compute, sp_ids), headers=["molecule id", "singlepoint id"])

        if not (opt_mols_to_compute or sp_mols_to_compute):
            output = "\n\n" + "*" * 80 + "\n"
            output += "All reaction components are complete!\n\n"

            output += "Reaction results:\n"
            table = []
            total_energy = 0.0

            coef_map = {x.molecule_id: x.coefficient for x in rxn_orm.components}

            for component in rxn_orm.components:
                mol_form = component.molecule.identifiers["molecular_formula"]
                mol_id = component.molecule_id
                coefficient = coef_map[mol_id]

                if has_optimization and has_singlepoint:
                    energy = component.singlepoint_record.properties["return_energy"]
                    table_row = [
                        mol_id,
                        mol_form,
                        component.optimization_record.final_molecule_id,
                        component.optimization_id,
                        component.singlepoint_id,
                        energy,
                        coefficient,
                    ]
                elif has_singlepoint:
                    energy = component.singlepoint_record.properties["return_energy"]
                    table_row = [mol_id, mol_form, component.singlepoint_id, energy, coefficient]
                else:
                    # has optimization only
                    energy = component.optimization_record.energies[-1]
                    table_row = [
                        mol_id,
                        mol_form,
                        component.optimization_record.final_molecule_id,
                        component.optimization_id,
                        energy,
                        coefficient,
                    ]

                table.append(table_row)
                total_energy += coefficient * energy

            if has_optimization and has_singlepoint:
                output += tabulate.tabulate(
                    table,
                    headers=[
                        "initial molecule id",
                        "molecule",
                        "optimized molecule id",
                        "optimization id",
                        "singlepoint id",
                        "energy (hartree)",
                        "coefficient",
                    ],
                )
            elif has_singlepoint:
                output += tabulate.tabulate(
                    table,
                    headers=["initial molecule id", "molecule", "singlepoint id", "energy (hartree)", "coefficient"],
                )
            else:
                output += tabulate.tabulate(
                    table,
                    headers=[
                        "initial molecule id",
                        "molecule",
                        "optimized molecule id",
                        "optimization id",
                        "energy (hartree)",
                        "coefficient",
                    ],
                )

            output += "\n\n"
            output += f"Total reaction energy: {total_energy:.16f} hartrees"

            rxn_orm.total_energy = total_energy

        append_output(session, rxn_orm, OutputTypeEnum.stdout, output)

        return not (opt_mols_to_compute or sp_mols_to_compute)

    def insert_full_qcportal_records_v1(
        self, session: Session, records: Sequence[ReactionRecord], creator_user_id: Optional[int]
    ) -> List[ReactionRecordORM]:
        ret = []

        rxn_specs = [r.specification for r in records]

        meta, spec_ids = self.root_socket.records.reaction.add_specifications(rxn_specs, session=session)
        if not meta.success:
            raise RuntimeError("Aborted reaction insertion - could not add specifications: " + meta.error_string)

        for record, spec_id in zip(records, spec_ids):
            record_orm = ReactionRecordORM(
                specification_id=spec_id,
                total_energy=record.total_energy,
                status=RecordStatusEnum.complete,
            )

            if record.component_records_:
                component_mols = [x.molecule for x in record.component_records_]
                component_sps = [x.singlepoint_record for x in record.component_records_]
                component_opts = [x.optimization_record for x in record.component_records_]

                assert len(component_mols) == len(component_sps) == len(component_opts)
                assert len(component_mols) > 0

                # All are none or all are not none
                assert all(x is None for x in component_sps) or all(x is not None for x in component_sps)
                assert all(x is None for x in component_opts) or all(x is not None for x in component_opts)

                meta, component_mol_ids = self.root_socket.molecules.add(component_mols, session=session)
                if not meta.success:
                    raise RuntimeError(
                        "Aborted reaction insertion - could not add initial molecules: " + meta.error_string
                    )

                if component_sps[0] is not None:
                    # all are not none
                    component_sp_ids = self.root_socket.records.insert_full_qcportal_records(
                        session, component_sps, creator_user_id
                    )
                else:
                    component_sp_ids = [None for _ in component_sps]

                if component_opts[0] is not None:
                    # all are not none
                    component_opt_ids = self.root_socket.records.insert_full_qcportal_records(
                        session, component_opts, creator_user_id
                    )
                else:
                    component_opt_ids = [None for _ in component_opts]

                record_orm.components = [
                    ReactionComponentORM(
                        molecule_id=mid,
                        coefficient=com.coefficient,
                        singlepoint_id=spid,
                        optimization_id=optid,
                    )
                    for mid, spid, optid, com in zip(
                        component_mol_ids, component_sp_ids, component_opt_ids, record.component_records_
                    )
                ]

            ret.append(record_orm)

        return ret

    def add_specifications(
        self, rxn_specs: Sequence[ReactionSpecification], *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, List[int]]:
        """
        Adds specifications for reaction services to the database, returning their IDs.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Specification IDs are returned in the same order as the input specifications

        Parameters
        ----------
        rxn_specs
            Sequence of specifications to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the IDs of the specifications.
        """

        to_add = []

        for rxn_spec in rxn_specs:
            kw_dict = rxn_spec.keywords.dict()

            rxn_spec_dict = {"program": rxn_spec.program, "keywords": kw_dict, "protocols": {}}
            rxn_spec_hash = hash_dict(rxn_spec_dict)

            rxn_spec_orm = ReactionSpecificationORM(
                program=rxn_spec.program,
                keywords=kw_dict,
                protocols=rxn_spec_dict["protocols"],
                specification_hash=rxn_spec_hash,
            )

            to_add.append(rxn_spec_orm)

        with self.root_socket.optional_session(session) as session:

            qc_specs_lst = [
                (idx, x.singlepoint_specification)
                for idx, x in enumerate(rxn_specs)
                if x.singlepoint_specification is not None
            ]
            opt_specs_lst = [
                (idx, x.optimization_specification)
                for idx, x in enumerate(rxn_specs)
                if x.optimization_specification is not None
            ]

            qc_specs_map = {}
            if qc_specs_lst:
                qc_specs = [x[1] for x in qc_specs_lst]
                meta, qc_spec_ids = self.root_socket.records.singlepoint.add_specifications(qc_specs, session=session)

                if not meta.success:
                    return (
                        InsertMetadata(
                            error_description="Unable to add singlepoint specifications: " + meta.error_string,
                        ),
                        [],
                    )

                qc_specs_map = {idx: qc_spec_id for (idx, _), qc_spec_id in zip(qc_specs_lst, qc_spec_ids)}

            opt_specs_map = {}
            if opt_specs_lst:
                opt_specs = [x[1] for x in opt_specs_lst]
                meta, opt_spec_ids = self.root_socket.records.optimization.add_specifications(
                    opt_specs, session=session
                )

                if not meta.success:
                    return (
                        InsertMetadata(
                            error_description="Unable to add optimization specifications: " + meta.error_string,
                        ),
                        [],
                    )

                opt_specs_map = {idx: opt_spec_id for (idx, _), opt_spec_id in zip(opt_specs_lst, opt_spec_ids)}

            for idx, rxn_spec_orm in enumerate(to_add):
                rxn_spec_orm.singlepoint_specification_id = qc_specs_map.get(idx, None)
                rxn_spec_orm.optimization_specification_id = opt_specs_map.get(idx, None)

            meta, ids = insert_general(
                session,
                to_add,
                (
                    ReactionSpecificationORM.specification_hash,
                    ReactionSpecificationORM.singlepoint_specification_id,
                    ReactionSpecificationORM.optimization_specification_id,
                ),
                (ReactionSpecificationORM.id,),
                use_unique=True,
            )

            return meta, [x[0] for x in ids]

    def add_specification(
        self, rxn_spec: ReactionSpecification, *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, Optional[int]]:
        """
        Adds a specification for a reaction service to the database, returning its id.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Parameters
        ----------
        rxn_spec
            Specification to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the id of the specification.
        """

        meta, ids = self.add_specifications([rxn_spec], session=session)

        if not ids:
            return meta, None

        return meta, ids[0]

    def get(
        self,
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        options = []

        if include:
            if is_included("components", include, exclude, False) or is_included("molecules", include, exclude, False):
                options.append(
                    selectinload(ReactionRecordORM.components).options(selectinload(ReactionComponentORM.molecule))
                )

        with self.root_socket.optional_session(session, True) as session:
            return self.root_socket.records.get_base(
                orm_type=self.record_orm,
                record_ids=record_ids,
                include=include,
                exclude=exclude,
                missing_ok=missing_ok,
                additional_options=options,
                session=session,
            )

    def query(
        self,
        query_data: ReactionQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        and_query = []
        need_qc_spec_join = False
        need_opt_spec_join = False
        need_spec_join = False
        need_component_join = False

        if query_data.program is not None:
            and_query.append(ReactionSpecificationORM.program.in_(query_data.program))
            need_spec_join = True
        if query_data.qc_program is not None:
            and_query.append(QCSpecificationORM.program.in_(query_data.qc_program))
            need_qc_spec_join = True
        if query_data.qc_method is not None:
            and_query.append(QCSpecificationORM.method.in_(query_data.qc_method))
            need_qc_spec_join = True
        if query_data.qc_basis is not None:
            and_query.append(QCSpecificationORM.basis.in_(query_data.qc_basis))
            need_qc_spec_join = True
        if query_data.optimization_program is not None:
            and_query.append(OptimizationSpecificationORM.program.in_(query_data.qc_basis))
            need_opt_spec_join = True
        if query_data.molecule_id is not None:
            and_query.append(ReactionComponentORM.molecule_id.in_(query_data.molecule_id))
            need_component_join = True

        stmt = select(ReactionRecordORM.id)

        if need_spec_join or need_qc_spec_join or need_opt_spec_join:
            stmt = stmt.join(ReactionRecordORM.specification)

        if need_qc_spec_join:
            stmt = stmt.join(ReactionSpecificationORM.singlepoint_specification)

        if need_opt_spec_join:
            stmt = stmt.join(ReactionSpecificationORM.optimization_specification)

        if need_component_join:
            # Do not load components as part of the ORM, but join for the query
            stmt = stmt.join(ReactionRecordORM.components)

        stmt = stmt.where(*and_query)

        return self.root_socket.records.query_base(
            stmt=stmt,
            orm_type=ReactionRecordORM,
            query_data=query_data,
            session=session,
        )

    def add_internal(
        self,
        stoichiometries: Sequence[Iterable[Tuple[float, int]]],  # coefficient, molecule_id
        rxn_spec_id: int,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Internal function for adding new reaction computations

        This function expects that the molecules and specification are already added to the
        database and that the ids are known.

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        stoichiometries
            Coefficients and IDs of the molecules that form the reaction
        rxn_spec_id
            ID of the specification
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user_id
            ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        compute_tag = compute_tag.lower()

        with self.root_socket.optional_session(session, False) as session:
            # Lock for the entire transaction
            session.execute(select(func.pg_advisory_xact_lock(reaction_insert_lock_id))).scalar()

            rxn_ids = []
            inserted_idx = []
            existing_idx = []

            if find_existing:
                # Deduplication is a bit complicated we have a many-to-many relationship
                # between reactions and molecules. So skip the general insert

                # Create a cte with the molecules we can query against
                # This is like a table, with the specification id and the molecule ids
                # as a postgres array (sorted)
                # We then use this to determine if there are duplicates
                init_mol_cte = (
                    select(
                        ReactionRecordORM.id,
                        ReactionRecordORM.specification_id,
                        array_agg(
                            aggregate_order_by(ReactionComponentORM.molecule_id, ReactionComponentORM.molecule_id.asc())
                        ).label("molecule_ids"),
                    )
                    .join(
                        ReactionComponentORM,
                        ReactionComponentORM.reaction_id == ReactionRecordORM.id,
                    )
                    .group_by(ReactionRecordORM.id)
                    .cte()
                )

                for idx, rxn_mols in enumerate(stoichiometries):
                    # sort molecule ids by increasing ids, and remove duplicates
                    rxn_mol_ids = sorted(set(x[1] for x in rxn_mols))

                    # does this exist?
                    stmt = select(init_mol_cte.c.id)
                    stmt = stmt.where(init_mol_cte.c.specification_id == rxn_spec_id)
                    stmt = stmt.where(init_mol_cte.c.molecule_ids == rxn_mol_ids)
                    existing = session.execute(stmt).scalars().first()

                    if not existing:
                        component_orm = [
                            ReactionComponentORM(coefficient=coeff, molecule_id=mid) for coeff, mid in rxn_mols
                        ]

                        rxn_orm = ReactionRecordORM(
                            is_service=True,
                            specification_id=rxn_spec_id,
                            components=component_orm,
                            status=RecordStatusEnum.waiting,
                            creator_user_id=creator_user_id,
                        )

                        self.create_service(rxn_orm, compute_tag, compute_priority, find_existing)

                        session.add(rxn_orm)
                        session.flush()

                        rxn_ids.append(rxn_orm.id)
                        inserted_idx.append(idx)
                    else:
                        rxn_ids.append(existing)
                        existing_idx.append(idx)

            else:  # don't find existing - always add
                for idx, rxn_mols in enumerate(stoichiometries):
                    # sort molecule ids by increasing ids, and remove duplicates
                    rxn_mol_ids = sorted(set(x[1] for x in rxn_mols))

                    component_orm = [
                        ReactionComponentORM(coefficient=coeff, molecule_id=mid) for coeff, mid in rxn_mols
                    ]

                    rxn_orm = ReactionRecordORM(
                        is_service=True,
                        specification_id=rxn_spec_id,
                        components=component_orm,
                        status=RecordStatusEnum.waiting,
                        creator_user_id=creator_user_id,
                    )

                    self.create_service(rxn_orm, compute_tag, compute_priority, find_existing)

                    session.add(rxn_orm)
                    session.flush()

                    rxn_ids.append(rxn_orm.id)
                    inserted_idx.append(idx)

            meta = InsertMetadata(inserted_idx=inserted_idx, existing_idx=existing_idx)
            return meta, rxn_ids

    def add(
        self,
        stoichiometries: Sequence[Iterable[Tuple[float, Union[int, Molecule]]]],
        rxn_spec: ReactionSpecification,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Adds new reaction calculations

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        stoichiometries
            Coefficient and molecules (objects or ids) to compute using the specification
        rxn_spec
            Specification for the reaction calculations
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user
            Name or ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input stoichiometries
        """

        with self.root_socket.optional_session(session, False) as session:
            creator_user_id = self.root_socket.users.get_optional_user_id(creator_user, session=session)

            # First, add the specification
            spec_meta, spec_id = self.add_specification(rxn_spec, session=session)

            if not spec_meta.success:
                return (
                    InsertMetadata(
                        error_description="Aborted - could not add specification: " + spec_meta.error_string
                    ),
                    [],
                )

            # Now the molecules
            new_mol = []
            for single_stoic in stoichiometries:
                mol_obj = [x[1] for x in single_stoic]
                mol_meta, mol_ids = self.root_socket.molecules.add_mixed(mol_obj, session=session)
                if not mol_meta.success:
                    return (
                        InsertMetadata(
                            error_description="Aborted - could not add all molecules: " + mol_meta.error_string
                        ),
                        [],
                    )

                new_mol.append([(x[0], y) for x, y in zip(single_stoic, mol_ids)])

            return self.add_internal(
                new_mol,
                spec_id,
                compute_tag,
                compute_priority,
                creator_user_id,
                find_existing,
                session=session,
            )

    def add_from_input(
        self,
        record_input: ReactionInput,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertCountsMetadata, int]:

        assert isinstance(record_input, ReactionInput)

        meta, ids = self.add(
            [record_input.stoichiometries],
            record_input.specification,
            compute_tag,
            compute_priority,
            creator_user,
            find_existing,
        )

        return InsertCountsMetadata.from_insert_metadata(meta), ids[0]

    ####################################################
    # Some stuff to be retrieved for reactions
    ####################################################

    def get_components(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        options = [
            lazyload("*"),
            defer("*"),
            joinedload(ReactionRecordORM.components).options(
                undefer("*"), joinedload(ReactionComponentORM.molecule).options(undefer("*"))
            ),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(ReactionRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")

            return [x.model_dict() for x in rec.components]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data, test_users
from qcfractal.components.neb.record_db_models import NEBRecordORM
from qcfractal.db_socket import SQLAlchemySocket
from qcfractal.testing_helpers import run_service
from qcportal.auth import UserInfo
from qcportal.neb import (
    NEBSpecification,
    NEBKeywords,
    compare_neb_records,
)
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import (
    QCSpecification,
    SinglepointProtocols,
)
from qcportal.utils import now_at_utc
from .testing_helpers import compare_neb_specs, test_specs, load_procedure_data, generate_task_key, load_record_data

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal.managers import ManagerName
    from sqlalchemy.orm.session import Session


@pytest.mark.parametrize("spec", test_specs)
def test_neb_socket_add_get(storage_socket: SQLAlchemySocket, session: Session, spec: NEBSpecification):
    chain1 = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]
    chain2 = [load_molecule_data("neb/neb_C3H2N_%i" % i) for i in range(21)]

    time_0 = now_at_utc()
    meta, ids = storage_socket.records.neb.add([chain1, chain2], spec, "tag1", PriorityEnum.low, None, True)
    time_1 = now_at_utc()
    assert meta.success

    recs = [session.get(NEBRecordORM, i) for i in ids]

    assert len(recs) == 2
    for r in recs:
        assert r.record_type == "neb"
        assert r.status == RecordStatusEnum.waiting
        assert compare_neb_specs(spec, r.specification.model_dict())

        # Service queue entry should exist with the proper tag and priority
        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    assert len(recs[0].initial_chain) == spec.keywords.images
    assert len(recs[1].initial_chain) == spec.keywords.images


def test_neb_socket_add_same_chains_diff_order(storage_socket: SQLAlchemySocket, session: Session):
    # Flipping the order of molecules in a chain generates different initial chain.
    spec = test_specs[0]
    chain1 = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]
    chain2 = chain1[::-1]

    # Now add records
    meta, ids = storage_socket.records.neb.add([chain1, chain2], spec, "*", PriorityEnum.normal, None, True)
    assert meta.success
    assert meta.n_inserted == 2
    assert meta.n_existing == 0

    recs = [session.get(NEBRecordORM, i) for i in ids]
    assert len(recs) == 2
    assert recs[0].id != recs[1].id


def test_neb_socket_add_same_1(storage_socket: SQLAlchemySocket):
    spec = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=11,
            spring_constant=0.1,
            spring_type=0,
        ),
        singlepoint_specification=QCSpecification(
            program="psi4",
            keywords={"k": "value"},
            driver="gradient",
            method="CCSD(T)",
            basis="def2-tzvp",
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    )
    chain1 = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]

    meta, id1 = storage_socket.records.neb.add([chain1], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.neb.add([chain1], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_neb_socket_add_same_2(storage_socket: SQLAlchemySocket):
    # some modifications to the input specification
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=11,
            spring_constant=0.1,
            spring_type=1,
        ),
        singlepoint_specification=QCSpecification(
            program="psi4",
            keywords={"k": "value"},
            driver="gradient",
            method="CCSD(T)",
            basis="def2-tzvp",
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    )

    spec2 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=11,
            spring_constant=0.1,
            spring_type=1,
        ),
        singlepoint_specification=QCSpecification(
            program="PSI4",
            keywords={"k": "value"},
            driver="gradient",
            method="ccsd(T)",
            basis="def2-tzvp",
            protocols=SinglepointProtocols(wavefunction="all", stdout=True),
        ),
    )

    chain1 = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]

    meta, id1 = storage_socket.records.neb.add([chain1], spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.neb.add([chain1, chain1, chain1], spec2, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 3
    assert meta.existing_idx == [0, 1, 2]
    assert id1[0] == id2[0]


def test_neb_socket_add_different_1(storage_socket: SQLAlchemySocket):
    # Molecules are a subset of another
    spec = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=11,
            spring_constant=0.1,
            spring_type=0,
        ),
        singlepoint_specification=QCSpecification(
            program="psi4",
            keywords={"k": "value"},
            driver="gradient",
            method="CCSD(T)",
            basis="def2-tzvp",
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    )
    chain1 = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]
    chain2 = [load_molecule_data("neb/neb_C3H2N_%i" % i) for i in range(21)]
    meta, id1 = storage_socket.records.neb.add([chain1, chain2], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 2
    assert meta.inserted_idx == [0, 1]

    meta, id2 = storage_socket.records.neb.add(
        [chain1, chain2, chain2[::-1]], spec, "*", PriorityEnum.normal, None, True
    )
    assert meta.n_inserted == 1
    assert meta.n_existing == 2
    assert meta.existing_idx == [0, 1]
    assert meta.inserted_idx == [2]
    assert id1[0] == id2[0]
    assert id1[1] == id2[1]


@pytest.mark.parametrize(
    "test_data_name",
    [
        "neb_HCN_psi4_pbe",
        "neb_HCN_psi4_pbe0_opt1",
        "neb_HCN_psi4_pbe_opt2",
        "neb_HCN_psi4_b3lyp_opt3",
        "neb_HCN_psi4_pbe_opt_diff",
    ],
)
def test_neb_socket_run(
    storage_socket: SQLAlchemySocket, session: Session, activated_manager_name: ManagerName, test_data_name: str
):
    input_spec_1, initial_chain_1, result_data_1 = load_procedure_data(test_data_name)

    storage_socket.users.add(UserInfo(username="submit_user", role="submit", enabled=True))

    meta_1, id_1 = storage_socket.records.neb.add(
        [initial_chain_1],
        input_spec_1,
        "test_tag",
        PriorityEnum.low,
        "submit_user",
        True,
    )
    assert meta_1.success

    time_0 = now_at_utc()
    finished, n_spopt = run_service(
        storage_socket, activated_manager_name, id_1[0], generate_task_key, result_data_1, 150
    )
    time_1 = now_at_utc()

    rec = session.get(NEBRecordORM, id_1)

    assert rec.status == RecordStatusEnum.complete
    assert time_0 < rec.modified_on < time_1
    assert len(rec.compute_history) == 1
    assert len(rec.compute_history[-1].outputs) == 1
    assert rec.compute_history[-1].status == RecordStatusEnum.complete
    assert time_0 < rec.compute_history[-1].modified_on < time_1
    assert rec.service is None

    desc_info = storage_socket.records.get_short_descriptions(id_1)[0]
    short_desc = desc_info["description"]
    assert desc_info["record_type"] == rec.record_type
    assert desc_info["created_on"] == rec.created_on
    assert rec.specification.program in short_desc
    assert rec.specification.singlepoint_specification.program in short_desc
    assert rec.specification.singlepoint_specification.method in short_desc

    out = rec.compute_history[-1].outputs["stdout"].get_output()
    assert "NEB calculation is completed" in out


def test_neb_socket_run_duplicate(
    storage_socket: SQLAlchemySocket,
    session: Session,
    activated_manager_name: ManagerName,
):
    input_spec_1, molecules_1, result_data_1 = load_procedure_data("neb_HCN_psi4_pbe_opt_diff")

    meta_1, id_1 = storage_socket.records.neb.add([molecules_1], input_spec_1, "test_tag", PriorityEnum.low, None, True)
    id_1 = id_1[0]
    assert meta_1.success

    run_service(storage_socket, activated_manager_name, id_1, generate_task_key, result_data_1, 100)

    rec_1 = session.get(NEBRecordORM, id_1)
    assert rec_1.status == RecordStatusEnum.complete
    sp_ids_1 = [x.singlepoint_id for x in rec_1.singlepoints]
    opt_ids_1 = [x.optimization_id for x in rec_1.optimizations]

    # Submit again, without duplicate checking
    meta_2, id_2 = storage_socket.records.neb.add(
        [molecules_1], input_spec_1, "test_tag", PriorityEnum.low, None, False
    )
    id_2 = id_2[0]
    assert meta_2.success
    assert id_2 != id_1

    run_service(storage_socket, activated_manager_name, id_2, generate_task_key, result_data_1, 1000)

    rec_2 = session.get(NEBRecordORM, id_2)
    assert rec_2.status == RecordStatusEnum.complete
    sp_ids_2 = [x.singlepoint_id for x in rec_2.singlepoints]
    opt_ids_2 = [x.optimization_id for x in rec_2.optimizations]

    assert set(sp_ids_1 + opt_ids_1).isdisjoint(sp_ids_2 + opt_ids_2)


def test_neb_socket_insert_full_qcportal_record(secure_snowflake: QCATestingSnowflake):
    test_names = [
        "neb_HCN_psi4_b3lyp_opt3",
        "neb_HCN_psi4_pbe0_opt1",
        "neb_HCN_psi4_pbe",
        "neb_HCN_psi4_pbe_opt2",
        "neb_HCN_psi4_pbe_opt_diff",
        "neb_error_68251163",
    ]

    storage_socket = secure_snowflake.get_storage_socket()
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    user_id = client.get_user().id

    for test_name in test_names:
        initial_record = load_record_data(test_name)
        initial_record_copy = initial_record.copy(deep=True)

        # Need a full copy of results - they can get mutated
        with storage_socket.session_scope() as session:
            ins_ids = storage_socket.records.insert_full_qcportal_records(session, [initial_record_copy], user_id)

        rec_1 = client.get_nebs(ins_ids[0], include=["**"])
        rec_1.fetch_children(include=["**"], force_fetch=True)
        compare_neb_records(rec_1, initial_record)

        assert rec_1.creator_user == "submit_user"
        for sps in rec_1.singlepoints.values():
            for sp in sps:
                assert sp.creator_user == "submit_user"
        for opt in rec_1.optimizations.values():
            assert opt.creator_user == "submit_user"

        if rec_1.ts_hessian:
            assert rec_1.ts_hessian.creator_user == "submit_user"



"""
Tests the tasks socket with respect to misbehaving managers
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import caplog_handler_at_level
from qcfractal.components.managers.db_models import ComputeManagerORM
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.testing_helpers import load_procedure_data, submit_procedure_data
from qcfractalcompute.compress import compress_result
from qcportal.exceptions import ComputeManagerError
from qcportal.managers import ManagerName
from qcportal.record_models import RecordStatusEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcfractal.db_socket import SQLAlchemySocket
    from sqlalchemy.orm.session import Session

_manager_programs = {"qcengine": ["unknown"], "psi4": ["unknown"], "qchem": ["v3.0"]}


def test_task_socket_claim_manager_noexist(storage_socket: SQLAlchemySocket):
    # Manager that doesn't exist tries to claim tasks

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    with pytest.raises(ComputeManagerError, match="does not exist") as err:
        storage_socket.tasks.claim_tasks(mname1.fullname, _manager_programs, ["tag1"])


def test_task_socket_claim_manager_inactive(storage_socket: SQLAlchemySocket):
    # Manager that is inactive tries to claim tasks

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=_manager_programs,
        compute_tags=["tag1"],
    )

    storage_socket.managers.deactivate([mname1.fullname])

    with pytest.raises(ComputeManagerError, match="is not active"):
        storage_socket.tasks.claim_tasks(mname1.fullname, _manager_programs, ["tag1"])


def test_task_socket_return_manager_noexist(storage_socket: SQLAlchemySocket, session: Session):
    # Manager that returns data does not exist

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=_manager_programs,
        compute_tags=["tag1"],
    )

    record_id, result_data = submit_procedure_data(storage_socket, "sp_psi4_benzene_energy_1", "tag1")
    result_data_compressed = compress_result(result_data.dict())

    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, _manager_programs, ["tag1"])

    with pytest.raises(ComputeManagerError, match="does not exist"):
        storage_socket.tasks.update_finished(
            "missing_manager",
            {tasks[0]["id"]: result_data_compressed},
        )

    # Task should still be running
    sp_rec = session.get(BaseRecordORM, record_id)
    assert sp_rec.status == RecordStatusEnum.running
    assert sp_rec.manager_name == mname1.fullname
    assert sp_rec.task is not None
    assert sp_rec.compute_history == []


def test_task_socket_return_manager_inactive(storage_socket: SQLAlchemySocket):
    # Manager that returns data does not exist

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=_manager_programs,
        compute_tags=["tag1"],
    )

    record_id, result_data = submit_procedure_data(storage_socket, "sp_psi4_benzene_energy_1", "tag1")
    result_data_compressed = compress_result(result_data.dict())

    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, _manager_programs, ["tag1"])

    storage_socket.managers.deactivate([mname1.fullname])

    with pytest.raises(ComputeManagerError, match="is not active"):
        storage_socket.tasks.update_finished(
            mname1.fullname,
            {tasks[0]["id"]: result_data_compressed},
        )


def test_task_socket_return_wrongmanager(storage_socket: SQLAlchemySocket, session: Session):
    # Manager returns data for a record that it hasn't claimed (or was stolen from it)

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    mname2 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="2345-6789-0123-4567")
    mid_1 = storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "psi4": ["unknown"], "qchem": ["v3.0"]},
        compute_tags=["tag1"],
    )

    mid_2 = storage_socket.managers.activate(
        name_data=mname2,
        manager_version="v2.0",
        username="bill",
        programs=_manager_programs,
        compute_tags=["tag1"],
    )

    record_id, result_data = submit_procedure_data(storage_socket, "sp_psi4_benzene_energy_1", "tag1")
    result_data_compressed = compress_result(result_data.dict())

    # Manager 1 claims tasks
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, _manager_programs, ["tag1"])

    # Manager 2 tries to return it
    rmeta = storage_socket.tasks.update_finished(
        mname2.fullname,
        {tasks[0]["id"]: result_data_compressed},
    )

    assert rmeta.n_accepted == 0
    assert rmeta.n_rejected == 1
    assert rmeta.rejected_info[0][0] == tasks[0]["id"]
    assert rmeta.rejected_info[0][1] == "Task is claimed by another manager"

    # But it didn't do anything
    # Task should still be running
    sp_rec = session.get(BaseRecordORM, record_id)
    assert sp_rec.status == RecordStatusEnum.running
    assert sp_rec.manager_name == mname1.fullname
    assert sp_rec.task is not None
    assert sp_rec.compute_history == []

    # Make sure manager info was updated
    manager = session.get(ComputeManagerORM, mid_2)
    assert manager.successes == 0
    assert manager.failures == 0
    assert manager.rejected == 1


def test_task_socket_return_manager_badid(snowflake: QCATestingSnowflake, caplog):
    storage_socket = snowflake.get_storage_socket()

    mname, mid = snowflake.activate_manager()

    # Manager returns data for a record that doesn't exist

    _, _, result_data = load_procedure_data("sp_psi4_benzene_energy_1")
    result_data_compressed = compress_result(result_data.dict())

    # Should be logged
    with caplog_handler_at_level(caplog, logging.WARNING):
        rmeta = storage_socket.tasks.update_finished(mname.fullname, {123: result_data_compressed})
        assert "does not exist in the task queue" in caplog.text

    assert rmeta.n_accepted == 0
    assert rmeta.n_rejected == 1
    assert rmeta.rejected_info[0][0] == 123
    assert rmeta.rejected_info[0][1] == "Task does not exist in the task queue"

    # Make sure manager info was updated
    with storage_socket.session_scope() as session:
        manager = session.get(ComputeManagerORM, mid)
        assert manager.successes == 0
        assert manager.failures == 0
        assert manager.rejected == 1


def test_task_socket_return_manager_badstatus_1(storage_socket: SQLAlchemySocket, session: Session, caplog):
    # Manager returns data for a record that is not running

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    mid = storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=_manager_programs,
        compute_tags=["tag1"],
    )

    record_id, result_data = submit_procedure_data(storage_socket, "sp_psi4_benzene_energy_1", "tag1")
    result_data_compressed = compress_result(result_data.dict())

    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, _manager_programs, ["tag1"])

    storage_socket.records.reset_running([record_id])

    with caplog_handler_at_level(caplog, logging.WARNING):
        rmeta = storage_socket.tasks.update_finished(mname1.fullname, {tasks[0]["id"]: result_data_compressed})
        assert "not in a running state" in caplog.text

    assert rmeta.n_accepted == 0
    assert rmeta.n_rejected == 1
    assert rmeta.rejected_info[0][0] == tasks[0]["id"]
    assert rmeta.rejected_info[0][1] == "Task is not in a running state"

    # Record should still be waiting
    sp_rec = session.get(BaseRecordORM, record_id)
    assert sp_rec.status == RecordStatusEnum.waiting
    assert sp_rec.manager_name is None
    assert sp_rec.task is not None
    assert sp_rec.compute_history == []

    # Make sure manager info was updated
    manager = session.get(ComputeManagerORM, mid)
    assert manager.successes == 0
    assert manager.failures == 0
    assert manager.rejected == 1


def test_task_socket_return_manager_badstatus_2(storage_socket: SQLAlchemySocket, session: Session, caplog):
    # Manager returns data for a record that completed (and therefore not in the task queue)

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    mid = storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=_manager_programs,
        compute_tags=["tag1"],
    )

    record_id, result_data = submit_procedure_data(storage_socket, "sp_psi4_benzene_energy_1", "tag1")
    result_data_compressed = compress_result(result_data.dict())

    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, _manager_programs, ["tag1"])

    storage_socket.tasks.update_finished(
        mname1.fullname,
        {tasks[0]["id"]: result_data_compressed},
    )

    time_1 = now_at_utc()

    with caplog_handler_at_level(caplog, logging.WARNING):
        rmeta = storage_socket.tasks.update_finished(
            mname1.fullname,
            {tasks[0]["id"]: result_data_compressed},
        )
        assert "does not exist in the task queue" in caplog.text

    assert rmeta.n_accepted == 0
    assert rmeta.n_rejected == 1
    assert rmeta.rejected_info[0][0] == tasks[0]["id"]
    assert rmeta.rejected_info[0][1] == "Task does not exist in the task queue"

    # Record should be complete
    sp_rec = session.get(BaseRecordORM, record_id)
    assert sp_rec.status == RecordStatusEnum.complete
    assert sp_rec.manager_name == mname1.fullname
    assert sp_rec.task is None
    assert len(sp_rec.compute_history) == 1
    assert sp_rec.modified_on < time_1

    # Make sure manager info was updated
    manager = session.get(ComputeManagerORM, mid)
    assert manager.successes == 1  # from the first submission
    assert manager.failures == 0
    assert manager.rejected == 1


def test_task_socket_return_manager_badstatus_3(storage_socket: SQLAlchemySocket, session: Session, caplog):
    # Manager returns data for a record that is cancelled

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    mid = storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=_manager_programs,
        compute_tags=["tag1"],
    )

    record_id, result_data = submit_procedure_data(storage_socket, "sp_psi4_benzene_energy_1", "tag1")
    result_data_compressed = compress_result(result_data.dict())

    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, _manager_programs, ["tag1"])

    storage_socket.records.cancel([record_id])

    time_1 = now_at_utc()

    with caplog_handler_at_level(caplog, logging.WARNING):
        rmeta = storage_socket.tasks.update_finished(
            mname1.fullname,
            {tasks[0]["id"]: result_data_compressed},
        )
        assert "does not exist in the task queue" in caplog.text

    assert rmeta.n_accepted == 0
    assert rmeta.n_rejected == 1
    assert rmeta.rejected_info[0][0] == tasks[0]["id"]
    assert rmeta.rejected_info[0][1] == "Task does not exist in the task queue"

    # Record should be cancelled
    sp_rec = session.get(BaseRecordORM, record_id)
    assert sp_rec.status == RecordStatusEnum.cancelled
    assert sp_rec.manager_name is None
    assert sp_rec.task is None
    assert len(sp_rec.compute_history) == 0
    assert sp_rec.modified_on < time_1

    # Make sure manager info was updated
    manager = session.get(ComputeManagerORM, mid)
    assert manager.successes == 0
    assert manager.failures == 0
    assert manager.rejected == 1



from __future__ import annotations

from typing import Optional, Type, Union, Tuple, Sequence, List, Dict, Any

from sqlalchemy import select, func
from sqlalchemy.orm import Session, lazyload, defer, joinedload, undefer, defaultload

from qcfractal.components.record_db_models import BaseRecordORM, OutputStoreORM, RecordComputeHistoryORM, NativeFileORM
from qcfractal.components.record_db_views import RecordDirectChildrenView
from qcfractal.components.services.db_models import ServiceQueueORM
from qcfractal.components.tasks.db_models import TaskQueueORM
from qcfractal.db_socket import SQLAlchemySocket
from qcportal.all_results import AllResultTypes
from qcportal.compression import CompressionEnum, decompress
from qcportal.exceptions import MissingDataError
from qcportal.metadata_models import InsertCountsMetadata
from qcportal.record_models import PriorityEnum, RecordStatusEnum, RecordQueryFilters, OutputTypeEnum


class BaseRecordSocket:
    """
    Base class for all record sockets
    """

    # Must be overridden by derived classes
    record_orm: Optional[Type[BaseRecordORM]] = None

    # Overridden for most classes, but not all
    record_input_type = None
    record_multi_input_type = None

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket

        # Make sure these were set by the derived classes
        assert self.record_orm is not None

    @staticmethod
    def create_task(record_orm: BaseRecordORM, compute_tag: str, compute_priority: PriorityEnum) -> None:
        """
        Create an entry in the task queue, and attach it to the given record ORM
        """

        available = record_orm.status == RecordStatusEnum.waiting
        record_orm.task = TaskQueueORM(
            compute_tag=compute_tag,
            compute_priority=compute_priority,
            required_programs=record_orm.required_programs,
            available=available,
        )

    @staticmethod
    def create_service(
        record_orm: BaseRecordORM, compute_tag: str, compute_priority: PriorityEnum, find_existing: bool
    ) -> None:
        """
        Create an entry in the service queue, and attach it to the given record ORM
        """

        record_orm.service = ServiceQueueORM(
            service_state={}, compute_tag=compute_tag, compute_priority=compute_priority, find_existing=find_existing
        )

    def add_from_input(
        self,
        record_input,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertCountsMetadata, int]:
        """
        Adds a record given an object of the record's input type

        Parameters
        ----------
        record_input
            The input specifying the calculation. Dependent on the type of record
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user
            Name or ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the ID of the record
        """

        raise NotImplementedError(
            f"add_from_input function not implemented for {type(self)}! This is a developer error"
        )

    def get(
        self,
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        """
        Obtain a record with specified IDs

        This function should be usable with all sockets - it uses the ORM type that was given
        in the constructor.

        Parameters
        ----------
        record_ids
            A list or other sequence of record IDs
        include
            Which fields of the result to return. Default is to return all fields.
        exclude
            Remove these fields from the return. Default is to return all fields.
        missing_ok
           If set to True, then missing results will be tolerated, and the returned list of
           records will contain None for the corresponding IDs that were not found.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Records as a dictionary in the same order as the given ids.
            If missing_ok is True, then this list will contain None where the molecule was missing.
        """

        raise NotImplementedError(f"get function not implemented for {type(self)}! This is a developer error")

    def query(
        self,
        query_data: RecordQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        """
        Query records of a particular type

        Parameters
        ----------
        query_data
            Fields/filters to query for
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of record ids that were found in the database.
        """

        raise NotImplementedError(f"query function not implemented for {type(self)}! This is a developer error")

    def generate_task_specifications(self, session: Session, record_ids: Sequence[int]) -> List[Dict[str, Any]]:
        """
        Generate the actual QCSchema input and related fields for a task
        """

        raise NotImplementedError(
            f"generate_task_specifications not implemented for {type(self)}! This is a developer error"
        )

    def update_completed_schema_v1(self, session: Session, record_id: int, result: AllResultTypes) -> None:
        """
        Update a record ORM based on the result of a successfully-completed computation
        """
        raise NotImplementedError(
            f"update_completed_schema_v1 not implemented for {type(self)}! This is a developer error"
        )

    def insert_full_qcportal_records_v1(
        self, session: Session, records: Sequence[AllResultTypes], creator_user_id: Optional[int]
    ) -> List[BaseRecordORM]:
        """
        Insert records into the database from a QCSchema result

        This will always create new ORMs from scratch, and not update any existing records.
        """
        raise NotImplementedError(
            f"insert_full_qcportal_records_v1 not implemented for {type(self)}! This is a developer error"
        )

    def insert_full_schema_v1(
        self, session: Session, results: Sequence[AllResultTypes], creator_user_id: Optional[int]
    ) -> List[BaseRecordORM]:
        """
        Insert records into the database from a QCSchema result

        This will always create new ORMs from scratch, and not update any existing records.
        """
        raise NotImplementedError(f"insert_full_schema_v1 not implemented for {type(self)}! This is a developer error")

    def initialize_service(self, session: Session, service_orm: ServiceQueueORM) -> None:
        """
        Initialize a new service

        A service is initialized when it moves from a waiting state to a running state. After it is
        initialized, iterate_service may be called.
        """
        raise NotImplementedError(f"initialize_service not implemented for {type(self)}! This is a developer error")

    def iterate_service(self, session: Session, service_orm: ServiceQueueORM) -> bool:
        """
        Iterate a service with successfully-completed dependencies

        The server will check to see if all dependencies are successfully completed. If they are, this function
        is called.

        Cases where the service has an errored dependency are handled elsewhere.
        """
        raise NotImplementedError(f"iterate_service not implemented for {type(self)}! This is a developer error")

    def available(self) -> bool:
        """
        Returns True if this is not a service, or if it is a service and available for iteration

        A service may not be available for iteration if the proper packages aren't installed
        on the server.
        """

        # By default, return True. Should be overridden by services
        return True

    ###########################################################################################
    # Getting various fields
    # The ones here apply to all records
    ###########################################################################################
    def get_comments(self, record_id: int, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        options = [
            lazyload("*"),
            defer("*"),
            joinedload(BaseRecordORM.comments).options(undefer("*"), defaultload("*")),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(BaseRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.model_dict() for x in rec.comments]

    def get_task(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Optional[Dict[str, Any]]:
        options = [lazyload("*"), defer("*"), joinedload(BaseRecordORM.task).options(undefer("*"), defaultload("*"))]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(BaseRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            if rec.task is None:
                return None
            return rec.task.model_dict()

    def get_service(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Optional[Dict[str, Any]]:
        options = [
            lazyload("*"),
            defer("*"),
            joinedload(BaseRecordORM.service).options(
                undefer("*"), defaultload("*"), joinedload(ServiceQueueORM.dependencies).options(undefer("*"))
            ),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(BaseRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            if rec.service is None:
                return None
            return rec.service.model_dict()

    def get_all_compute_history(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        options = [
            lazyload("*"),
            defer("*"),
            joinedload(BaseRecordORM.compute_history).options(undefer("*"), defaultload("*")),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(BaseRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.model_dict() for x in rec.compute_history]

    def get_single_compute_history(
        self,
        record_id: int,
        history_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Any]:
        # Slightly inefficient, but should be rarely called
        histories = self.get_all_compute_history(record_id, session=session)

        for h in histories:
            if h["id"] == history_id:
                return h
        else:
            raise MissingDataError(f"Cannot find compute history {history_id} for record {record_id}")

    def get_all_output_metadata(
        self,
        record_id: int,
        history_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Dict[str, Any]]:
        # Outer join with the history and record orm table to ensure ids, types, etc, match
        stmt = select(self.record_orm.id, OutputStoreORM)
        stmt = stmt.join(RecordComputeHistoryORM, self.record_orm.id == RecordComputeHistoryORM.record_id, isouter=True)
        stmt = stmt.join(OutputStoreORM, OutputStoreORM.history_id == RecordComputeHistoryORM.id, isouter=True)
        stmt = stmt.where(RecordComputeHistoryORM.id == history_id)
        stmt = stmt.where(self.record_orm.id == record_id)

        with self.root_socket.optional_session(session, True) as session:
            outputs = session.execute(stmt).all()

            # if empty list, not even the record was found
            if len(outputs) == 0:
                raise MissingDataError(f"Cannot find record {record_id}")

            # if the length is one, but the history is None, then record found, but no history
            elif len(outputs) == 1 and outputs[0][1] is None:
                return {}

            # Otherwise, found some outputs that match the record id/type
            else:
                return {o[1].output_type: o[1].model_dict() for o in outputs}

    def get_single_output_metadata(
        self,
        record_id: int,
        history_id: int,
        output_type: str,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Any]:
        # Slightly inefficient, but should be ok
        all_outputs = self.get_all_output_metadata(record_id, history_id, session=session)

        if output_type not in all_outputs:
            raise MissingDataError(f"Cannot find output {output_type} for record {record_id}/history {history_id}")

        return all_outputs[output_type]

    def get_single_output_rawdata(
        self,
        record_id: int,
        history_id: int,
        output_type: str,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[bytes, CompressionEnum]:
        stmt = select(OutputStoreORM.data, OutputStoreORM.compression_type)
        stmt = stmt.join(RecordComputeHistoryORM, RecordComputeHistoryORM.id == OutputStoreORM.history_id)
        stmt = stmt.join(self.record_orm, RecordComputeHistoryORM.record_id == self.record_orm.id)
        stmt = stmt.where(RecordComputeHistoryORM.record_id == record_id)
        stmt = stmt.where(OutputStoreORM.history_id == history_id)
        stmt = stmt.where(OutputStoreORM.output_type == output_type)

        with self.root_socket.optional_session(session, True) as session:
            output_data = session.execute(stmt).one_or_none()
            if output_data is None:
                raise MissingDataError(
                    f"Record {record_id}/history {history_id} does not have {output_type} output (or record/history does not exist)"
                )

            return output_data[0], output_data[1]

    def get_single_output_uncompressed(
        self, record_id: int, history_id: int, output_type: OutputTypeEnum, *, session: Optional[Session] = None
    ) -> Any:
        """
        Get an uncompressed output from a record
        """

        raw_data, ctype = self.get_single_output_rawdata(record_id, history_id, output_type, session=session)
        return decompress(raw_data, ctype)

    def get_all_native_files_metadata(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Dict[str, Any]]:
        options = [
            lazyload("*"),
            defer("*"),
            joinedload(BaseRecordORM.native_files).options(undefer("*"), defaultload("*")),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(BaseRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return {k: v.model_dict() for k, v in rec.native_files.items()}

    def get_single_native_file_metadata(
        self,
        record_id: int,
        name: str,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Any]:
        # Slightly inefficient, but should be ok
        all_nf = self.get_all_native_files_metadata(record_id, session=session)

        if not name in all_nf:
            raise MissingDataError(f"Cannot find native file {name} for record {record_id}")
        else:
            return all_nf[name]

    def get_single_native_file_rawdata(
        self,
        record_id: int,
        name: str,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[bytes, CompressionEnum]:
        stmt = select(NativeFileORM.data, NativeFileORM.compression_type)
        stmt = stmt.join(self.record_orm, NativeFileORM.record_id == self.record_orm.id)
        stmt = stmt.where(NativeFileORM.name == name)

        with self.root_socket.optional_session(session, True) as session:
            nf_data = session.execute(stmt).one_or_none()
            if nf_data is None:
                raise MissingDataError(
                    f"Record {record_id} does not have native file {name} (or record does not exist)"
                )

            return nf_data[0], nf_data[1]

    def get_single_native_file_uncompressed(
        self, record_id: int, name: str, *, session: Optional[Session] = None
    ) -> Any:
        """
        Get an uncompressed output from a record
        """

        raw_data, ctype = self.get_single_native_file_rawdata(record_id, name, session=session)
        return decompress(raw_data, ctype)

    def get_children_status(self, record_id: int, *, session: Optional[Session] = None) -> Dict[RecordStatusEnum, int]:
        stmt = select(BaseRecordORM.status, func.count())
        stmt = stmt.join(RecordDirectChildrenView, RecordDirectChildrenView.c.child_id == BaseRecordORM.id)
        stmt = stmt.where(RecordDirectChildrenView.c.parent_id == record_id)
        stmt = stmt.group_by(BaseRecordORM.status)

        with self.root_socket.optional_session(session, True) as session:
            res = session.execute(stmt).all()
            return {x: y for x, y in res}

    def get_children_errors(self, record_id: int, *, session: Optional[Session] = None) -> List[int]:
        stmt = select(RecordDirectChildrenView.c.child_id).distinct(RecordDirectChildrenView.c.child_id)
        stmt = stmt.join(BaseRecordORM, BaseRecordORM.id == RecordDirectChildrenView.c.child_id)
        stmt = stmt.where(RecordDirectChildrenView.c.parent_id == record_id)
        stmt = stmt.where(BaseRecordORM.status == RecordStatusEnum.error)

        with self.root_socket.optional_session(session, True) as session:
            return session.execute(stmt).scalars().all()



"""
Tests the tasks socket (claiming & returning data)
"""

from __future__ import annotations

import re

from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcfractal.components.optimization.testing_helpers import load_procedure_data as load_opt_procedure_data
from qcfractal.components.singlepoint.testing_helpers import (
    load_procedure_data as load_sp_procedure_data,
)
from qcfractal.components.torsiondrive.testing_helpers import load_procedure_data as load_td_procedure_data
from qcportal.managers import ManagerName
from qcportal.record_models import PriorityEnum


def test_record_client_waiting_reason(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    input_spec_1, molecule_1, result_data_1 = load_sp_procedure_data("sp_psi4_water_energy")
    input_spec_2, molecule_2, result_data_2 = load_opt_procedure_data("opt_psi4_benzene")
    input_spec_3, molecule_3, result_data_3 = load_sp_procedure_data("sp_rdkit_benzene_energy")

    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "tag1", PriorityEnum.low, None, True
    )
    meta, id_2 = storage_socket.records.optimization.add(
        [molecule_2], input_spec_2, "tag2", PriorityEnum.normal, None, True
    )
    meta, id_3 = storage_socket.records.singlepoint.add(
        [molecule_3], input_spec_3, "tag3", PriorityEnum.high, None, True
    )
    id_1 = id_1[0]
    id_2 = id_2[0]
    id_3 = id_3[0]
    all_id = [id_1, id_2, id_3]

    for i in all_id:
        reason = snowflake_client.get_waiting_reason(i)
        assert reason["reason"] == "No active managers"

    # Activate some managers
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "other_prog": ["unknown"]},
        compute_tags=["tag1"],
    )

    mname2 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-7890")
    storage_socket.managers.activate(
        name_data=mname2,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "geometric": ["unknown"]},
        compute_tags=["tag2"],
    )

    mname3 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-8888")
    storage_socket.managers.activate(
        name_data=mname3,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "psi4": ["unknown"]},
        compute_tags=["tag999"],
    )

    mname4 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-0123")
    storage_socket.managers.activate(
        name_data=mname4,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["unknown"]},
        compute_tags=["tag999"],
    )

    reason = snowflake_client.get_waiting_reason(id_1)
    assert reason["reason"] == "No manager matches programs & tags"
    assert re.search(r"missing programs.*psi4", reason["details"][mname1.fullname])
    assert re.search(r"missing programs.*psi4", reason["details"][mname2.fullname])
    assert re.search(r"does not handle tag.*tag1", reason["details"][mname3.fullname])
    assert re.search(r"does not handle tag.*tag1", reason["details"][mname4.fullname])

    reason = snowflake_client.get_waiting_reason(id_2)
    assert reason["reason"] == "No manager matches programs & tags"
    assert re.search(r"missing programs.*psi4", reason["details"][mname1.fullname])
    assert re.search(r"missing programs.*geometric", reason["details"][mname1.fullname])
    assert re.search(r"missing programs.*psi4", reason["details"][mname2.fullname])
    assert re.search(r"missing programs.*geometric", reason["details"][mname3.fullname])
    assert re.search(r"does not handle tag.*tag2", reason["details"][mname4.fullname])

    # Add a working manager
    mname5 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-2222")
    storage_socket.managers.activate(
        name_data=mname5,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["unknown"]},
        compute_tags=["tag1", "tag2"],
    )

    reason = snowflake_client.get_waiting_reason(id_1)
    assert reason["reason"] == "Waiting for a free manager"
    assert reason["details"][mname5.fullname] == "Manager is busy"

    reason = snowflake_client.get_waiting_reason(id_2)
    assert reason["reason"] == "Waiting for a free manager"
    assert reason["details"][mname5.fullname] == "Manager is busy"

    # third test record requires rdkit, which we haven't given yet
    reason = snowflake_client.get_waiting_reason(id_3)
    assert reason["reason"] == "No manager matches programs & tags"
    assert re.search(r"missing programs.*rdkit", reason["details"][mname1.fullname])
    assert re.search(r"missing programs.*rdkit", reason["details"][mname2.fullname])
    assert re.search(r"missing programs.*rdkit", reason["details"][mname3.fullname])
    assert re.search(r"missing programs.*rdkit", reason["details"][mname4.fullname])
    assert re.search(r"missing programs.*rdkit", reason["details"][mname5.fullname])

    # Also try through the record
    record = snowflake_client.get_records(id_3)
    assert record.get_waiting_reason() == reason

    # Add a working manager with * tag
    mname6 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-0011")
    storage_socket.managers.activate(
        name_data=mname6,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["unknown"], "rdkit": ["unknown"]},
        compute_tags=["tag1", "*"],
    )

    reason = snowflake_client.get_waiting_reason(id_3)
    assert reason["reason"] == "Waiting for a free manager"
    assert re.search(r"Manager is busy", reason["details"][mname6.fullname])


def test_record_client_waiting_reason_2(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    input_spec_1, molecule_1, result_data_1 = load_sp_procedure_data("sp_psi4_water_energy")
    input_spec_2, molecule_2, _ = load_td_procedure_data("td_H2O2_mopac_pm6")

    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "tag1", PriorityEnum.low, None, True
    )

    meta, id_2 = snowflake_client.add_torsiondrives(
        [molecule_2],
        "torsiondrive",
        keywords=input_spec_2.keywords,
        optimization_specification=input_spec_2.optimization_specification,
    )

    id_1 = id_1[0]
    id_2 = id_2[0]

    # Add a working manager with * tag
    mname6 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-0011")
    storage_socket.managers.activate(
        name_data=mname6,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["unknown"], "rdkit": ["unknown"]},
        compute_tags=["tag1", "*"],
    )

    # Should be able to be picked up
    reason = snowflake_client.get_waiting_reason(id_1)
    assert reason["reason"] == "Waiting for a free manager"
    assert re.search(r"Manager is busy", reason["details"][mname6.fullname])

    # Reason: not actually waiting
    storage_socket.records.cancel([id_1])
    reason = snowflake_client.get_waiting_reason(id_1)
    assert reason["reason"] == "Record is not waiting"

    # Reason: does not exist
    reason = snowflake_client.get_waiting_reason(id_1 + id_2)
    assert reason["reason"] == "Record does not exist"

    # Reason: is a service
    reason = snowflake_client.get_waiting_reason(id_2)
    assert reason["reason"] == "Record is a service"
